#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import operator
from dios import DictOfSeries
from typing import Callable
from saqc.constants import *
from saqc.core import register, Flags
from saqc.core.register import flagging
from saqc.lib.tools import statPass


@flagging()
def flagByStatLowPass(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    func: Callable[[np.array, pd.Series], float],
    window: str,
    thresh: float,
    sub_window: str = None,
    sub_thresh: float = None,
    min_periods: int = None,
    flag: float = BAD,
    **kwargs,
):
    """
    Flag *chunks* of length, `window`:

    1. If they excexceed `thresh` with regard to `stat`:
    2. If all (maybe overlapping) *sub-chunks* of *chunk*, with length `sub_window`,
       `excexceed `sub_thresh` with regard to `stat`:

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store quality flags to data.
    func: Callable[[np.array, pd.Series], float]
        Function to aggregate chunk contnent with.
    window: str
        Temporal extension of the chunks to test
    thresh: float
        Threshold, that triggers flagging, if exceeded by stat value.
    sub_window: str, default None,
        Window size of the sub chunks, that are additionally tested for exceeding
        `sub_thresh` with respect to `stat`.
    sub_thresh: float, default None
    min_periods: int, default None
    flag : float, default BAD
        flag to set

    Returns
    -------
    """

    datcol = data[field]
    if not min_periods:
        min_periods = 0
    if not sub_thresh:
        sub_thresh = thresh
    window = pd.Timedelta(window)

    if sub_window:
        sub_window = pd.Timedelta(sub_window)

    to_set = statPass(
        datcol, func, window, thresh, operator.gt, sub_window, sub_thresh, min_periods
    )
    flags[to_set, field] = flag
    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from typing import (
    Any,
    Callable,
    List,
    Sequence,
    Hashable,
    Tuple,
)
from copy import deepcopy, copy as shallowcopy

import pandas as pd
import numpy as np

from dios import DictOfSeries, to_dios
from saqc.constants import BAD

from saqc.core.modules import FunctionsMixin
from saqc.core.flags import initFlagsLike, Flags
from saqc.core.history import History
from saqc.core.register import FUNC_MAP, FunctionWrapper
from saqc.core.translation import (
    TranslationScheme,
    FloatScheme,
    SimpleScheme,
    PositionalScheme,
    DmpScheme,
)
from saqc.lib.tools import toSequence, concatDios
from saqc.lib.types import ExternalFlag, OptionalNone

# the import is needed to trigger the registration
# of the built-in (test-)functions
import saqc.funcs  # noqa

# warnings
pd.set_option("mode.chained_assignment", "warn")
np.seterr(invalid="ignore")


TRANSLATION_SCHEMES = {
    "float": FloatScheme,
    "simple": SimpleScheme,
    "dmp": DmpScheme,
    "positional": PositionalScheme,
}


class SaQC(FunctionsMixin):
    _attributes = {
        "_data",
        "_flags",
        "_scheme",
        "_attrs",
        "_called",
    }

    def __init__(
        self,
        data=None,
        flags=None,
        scheme: str | TranslationScheme = "float",
        copy: bool = True,
    ):
        self._data = self._initData(data, copy)
        self._flags = self._initFlags(flags, copy)
        self._scheme = self._initTranslationScheme(scheme)
        self._called = []
        self._attrs = {}
        self._validate(reason="init")

    def _construct(self, **attributes) -> SaQC:
        """
        Construct a new `SaQC`-Object from `self` and optionally inject
        attributes with any chechking and overhead.

        Parameters
        ----------
        **attributes: any of the `SaQC` data attributes with name and value

        Note
        ----
        For internal usage only! Setting values through `injectables` has
        the potential to mess up certain invariants of the constructed object.
        """
        out = SaQC(data=DictOfSeries(), flags=Flags(), scheme=self._scheme)
        out.attrs = self._attrs
        for k, v in attributes.items():
            if k not in self._attributes:
                raise AttributeError(f"SaQC has no attribute {repr(k)}")
            setattr(out, k, v)
        return out

    def _validate(self, reason=None):
        if not self._data.columns.equals(self._flags.columns):
            msg = "Consistency broken. data and flags have not the same columns."
            if reason:
                msg += f" This was most likely caused by: {reason}"
            raise RuntimeError(msg)

    @property
    def attrs(self) -> dict[Hashable, Any]:
        """
        Dictionary of global attributes of this dataset.
        """
        return self._attrs

    @attrs.setter
    def attrs(self, value: dict[Hashable, Any]) -> None:
        self._attrs = dict(value)

    @property
    def data_raw(self) -> DictOfSeries:
        return self._data

    @property
    def flags_raw(self) -> Flags:
        return self._flags

    @property
    def data(self) -> pd.DataFrame:
        data: pd.DataFrame = self._data.to_df()
        data.attrs = self._attrs.copy()
        return data

    @property
    def flags(self) -> pd.DataFrame:
        data: pd.DataFrame = self._scheme.backward(self._flags, attrs=self._attrs)
        data.attrs = self._attrs.copy()
        return data

    @property
    def result(self) -> SaQCResult:
        return SaQCResult(self._data, self._flags, self._attrs, self._scheme)

    def _expandFields(
        self,
        regex: bool,
        multivariate: bool,
        field: str | Sequence[str],
        target: str | Sequence[str] = None,
    ) -> Tuple[List[str], List[str]]:
        """
        check and expand `field` and `target`
        """

        if regex and target is not None:
            raise NotImplementedError(
                "explicit `target` not supported with regular expressions"
            )
        # expand regular expressions
        if regex:
            fmask = self._data.columns.str.match(field)
            fields = self._data.columns[fmask].tolist()
        else:
            fields = toSequence(field)

        targets = fields if target is None else toSequence(target)

        if multivariate:
            # wrap again to generalize the down stream loop
            fields, targets = [fields], [targets]
        else:
            if len(fields) != len(targets):
                raise ValueError(
                    "expected the same number of 'field' and 'target' values"
                )
        return fields, targets

    def _wrap(self, func: FunctionWrapper):
        """
        prepare user function input:
          - expand fields and targets
          - translate user given ``flag`` values or set the default ``BAD``
          - translate user given ``dfilter`` values or set the scheme default
          - dependeing on the workflow: initialize ``target`` variables

        Here we add the following parameters to all registered functions, regardless
        of their repsective definition:
          - ``regex``
          - ``target``

        """

        def inner(
            field: str | Sequence[str],
            *args,
            target: str | Sequence[str] = None,
            regex: bool = False,
            flag: ExternalFlag | OptionalNone = OptionalNone(),
            **kwargs,
        ) -> SaQC:

            kwargs.setdefault("dfilter", self._scheme.DFILTER_DEFAULT)

            if not isinstance(flag, OptionalNone):
                # translation schemes might want to use a flag,
                # None so we introduce a special class here
                kwargs["flag"] = self._scheme(flag)

            fields, targets = self._expandFields(
                regex=regex, multivariate=func.multivariate, field=field, target=target
            )

            out = self

            for field, target in zip(fields, targets):

                fkwargs = {
                    **kwargs,
                    "field": field,
                    "target": target,
                }

                if not func.handles_target and field != target:
                    if target not in self.data.columns:
                        out = out._callFunction(
                            FUNC_MAP["copyField"],
                            *args,
                            **fkwargs,
                        )
                    fkwargs["field"] = fkwargs.pop("target")

                out = out._callFunction(
                    func,
                    *args,
                    **fkwargs,
                )
            return out

        return inner

    def _callFunction(
        self,
        function: Callable,
        field: str | Sequence[str],
        *args: Any,
        **kwargs: Any,
    ) -> SaQC:

        res = function(data=self._data, flags=self._flags, field=field, *args, **kwargs)

        # keep consistence: if we modify data and flags inplace in a function,
        # but data is the original and flags is a copy (as currently implemented),
        # data and flags of the original saqc obj may change inconsistently.
        self._data, self._flags = res
        self._called += [(field, (function, args, kwargs))]
        self._validate(reason=f"call to {repr(function.__name__)}")

        return self._construct(
            _data=self._data, _flags=self._flags, _called=self._called
        )

    def __getattr__(self, key):
        """
        All failing attribute accesses are redirected to __getattr__.
        We use this mechanism to make the registered functions appear
        as `SaQC`-methods without actually implementing them.
        """
        if key not in FUNC_MAP:
            raise AttributeError(f"SaQC has no attribute {repr(key)}")
        return self._wrap(FUNC_MAP[key])

    def copy(self, deep=True):
        copyfunc = deepcopy if deep else shallowcopy
        new = self._construct()
        for attr in self._attributes:
            setattr(new, attr, copyfunc(getattr(self, attr)))
        return new

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memodict=None):
        return self.copy(deep=True)

    def _initTranslationScheme(
        self, scheme: str | TranslationScheme
    ) -> TranslationScheme:
        if isinstance(scheme, str) and scheme in TRANSLATION_SCHEMES:
            return TRANSLATION_SCHEMES[scheme]()
        if isinstance(scheme, TranslationScheme):
            return scheme
        raise TypeError(
            f"expected one of the following translation schemes '{TRANSLATION_SCHEMES.keys()} "
            f"or an initialized Translator object, got '{scheme}'"
        )

    def _initData(self, data, copy: bool) -> DictOfSeries:

        if data is None:
            return DictOfSeries()

        if isinstance(data, list):
            results = []
            for d in data:
                results.append(self._castToDios(d, copy=copy))
            return concatDios(results, warn=True, stacklevel=3)

        if isinstance(data, (DictOfSeries, pd.DataFrame, pd.Series)):
            return self._castToDios(data, copy)

        raise TypeError(
            "'data' must be of type pandas.Series, "
            "pandas.DataFrame or dios.DictOfSeries or"
            "a list of those."
        )

    def _castToDios(self, data, copy: bool):
        if isinstance(data, pd.Series):
            if not isinstance(data.name, str):
                raise ValueError(f"Cannot init from unnamed pd.Series")
            data = data.to_frame()
        if isinstance(data, pd.DataFrame):
            for idx in [data.index, data.columns]:
                if isinstance(idx, pd.MultiIndex):
                    raise TypeError("'data' should not have MultiIndex")
        data = to_dios(data)  # noop for DictOfSeries
        for c in data.columns:
            if not isinstance(c, str):
                raise TypeError("columns labels must be of type string")
        if copy:
            data = data.copy()
        return data

    def _initFlags(self, flags, copy: bool) -> Flags:
        if flags is None:
            return initFlagsLike(self._data)

        if isinstance(flags, list):
            result = Flags()
            for f in flags:
                f = self._castToFlags(f, copy=copy)
                for c in f.columns:
                    if c in result.columns:
                        warnings.warn(
                            f"Column {c} already exist. Data is overwritten. "
                            f"Avoid duplicate columns names over all inputs.",
                            stacklevel=2,
                        )
                        result.history[c] = f.history[c]
            flags = result

        elif isinstance(flags, (pd.DataFrame, DictOfSeries, Flags)):
            flags = self._castToFlags(flags, copy=copy)

        else:
            raise TypeError(
                "'flags' must be of type pandas.DataFrame, "
                "dios.DictOfSeries or saqc.Flags or "
                "a list of those."
            )

        # sanitize
        # - if column is missing flags but present in data, add it
        # - if column is present in both, the index must be equal
        for c in self._data.columns:
            if c not in flags.columns:
                flags.history[c] = History(self._data[c].index)
            else:
                if not flags[c].index.equals(self._data[c].index):
                    raise ValueError(
                        f"The flags index of column {c} does not equals "
                        f"the index of the same column in data."
                    )
        return flags

    def _castToFlags(self, flags, copy):
        if isinstance(flags, pd.DataFrame):
            for idx in [flags.index, flags.columns]:
                if isinstance(idx, pd.MultiIndex):
                    raise TypeError("'flags' should not have MultiIndex")
        if not isinstance(flags, Flags):
            flags = Flags(flags)
        if copy:
            flags = flags.copy()
        return flags


class SaQCResult:
    def __init__(
        self,
        data: DictOfSeries,
        flags: Flags,
        attrs: dict,
        scheme: TranslationScheme,
    ):
        assert isinstance(data, DictOfSeries)
        assert isinstance(flags, Flags)
        assert isinstance(attrs, dict)
        assert isinstance(scheme, TranslationScheme)
        self._data = data.copy()
        self._flags = flags.copy()
        self._attrs = attrs.copy()
        self._scheme = scheme
        self._validate()

        try:
            self._scheme.backward(self._flags, attrs=self._attrs)
        except Exception as e:
            raise RuntimeError("Translation of flags failed") from e

    def _validate(self):
        if not self._data.columns.equals(self._flags.columns):
            raise AssertionError(
                "Consistency broken. data and flags have not the same columns"
            )

    @property
    def data(self) -> pd.DataFrame:
        data: pd.DataFrame = self._data.copy().to_df()
        data.attrs = self._attrs.copy()
        return data

    @property
    def flags(self) -> pd.DataFrame:
        data: pd.DataFrame = self._scheme.backward(self._flags, attrs=self._attrs)
        data.attrs = self._attrs.copy()
        return data

    @property
    def data_raw(self) -> DictOfSeries:
        return self._data

    @property
    def flags_raw(self) -> Flags:
        return self._flags

    @property
    def columns(self) -> DictOfSeries():
        self._validate()
        return self._data.columns

    def __getitem__(self, key):
        self._validate()
        if key not in self.columns:
            raise KeyError(key)
        data_series = self._data[key].copy()
        # slice flags to one column
        flags = Flags({key: self._flags._data[key]}, copy=True)

        df = self._scheme.backward(flags, attrs=self._attrs)
        if isinstance(df.columns, pd.MultiIndex):
            df = df.droplevel(level=0, axis=1)

        if len(df.columns) == 1:
            df.columns = ["flags"]

        df.insert(0, column="data", value=data_series)
        df.columns.name = None
        df.index.name = None
        return df

    def __repr__(self):
        return f"SaQCResult\nColumns: {self.columns.to_list()}"



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import ast
from saqc.core.flags import Flags
from saqc.core.reader import fromConfig
import pytest
import numpy as np
import pandas as pd
import dios

from saqc.constants import *
from saqc.core import initFlagsLike
from saqc.core.visitor import ConfigFunctionParser
from saqc.core.register import register
from saqc.funcs.generic import _execGeneric
from saqc import SaQC

from tests.common import initData, writeIO


@pytest.fixture
def data():
    return initData()


@pytest.fixture
def data_diff():
    data = initData(cols=3)
    col0 = data[data.columns[0]]
    col1 = data[data.columns[1]]
    mid = len(col0) // 2
    offset = len(col0) // 8
    return dios.DictOfSeries(
        data={
            col0.name: col0.iloc[: mid + offset],
            col1.name: col1.iloc[mid - offset :],
        }
    )


def _compileGeneric(expr):
    tree = ast.parse(expr, mode="eval")
    _, kwargs = ConfigFunctionParser().parse(tree.body)
    return kwargs["func"]


def test_syntaxError():
    tests = [
        "range(x=5",
        "rangex=5)",
        "range[x=5]" "range{x=5}" "int->float(x=4)" "int*float(x=4)",
    ]

    for test in tests:
        with pytest.raises(SyntaxError):
            _compileGeneric(f"flag(func={test})")


def test_typeError():
    """
    test that forbidden constructs actually throw an error
    """

    # TODO: think about cases that should be forbidden
    tests = ("lambda x: x * 2",)

    for test in tests:
        with pytest.raises(TypeError):
            _compileGeneric(f"flagGeneric(func={test})")


def test_comparisonOperators(data):
    var1, var2, *_ = data.columns
    flags = initFlagsLike(data)

    tests = [
        (["var1"], "x > 100", data[var1] > 100),
        (["var2"], "10 >= y", 10 >= data[var2]),
        (["var2"], f"y < 100", data[var2] < 100),
        (["var1", "var2"], "x <= y", data[var1] <= data[var2]),
        (["var1", "var2"], "x == y", data[var1] == data[var2]),
        (["var1", "var2"], "x != y", data[var1] != data[var2]),
    ]

    for field, test, expected in tests:
        func = _compileGeneric(f"flagGeneric(func={test})")
        result = _execGeneric(Flags({f: flags[f] for f in field}), data[field], func)
        assert (result == expected).all(axis=None)


def test_arithmeticOperators(data):

    var1, *_ = data.columns

    data = data[var1]
    flags = Flags({var1: pd.Series(UNFLAGGED, index=data.index)})

    tests = [
        ("var1 + 100 > 110", data + 100 > 110),
        ("var1 - 100 > 0", data - 100 > 0),
        ("var1 * 100 > 200", data * 100 > 200),
        ("var1 / 100 > .1", data / 100 > 0.1),
        ("var1 % 2 == 1", data % 2 == 1),
        ("var1 ** 2 == 0", data ** 2 == 0),
    ]

    for test, expected in tests:
        func = _compileGeneric(f"processGeneric(func={test})")
        result = _execGeneric(flags, data, func)
        assert (result == expected).all(axis=None)


def test_nonReduncingBuiltins(data):
    var1, *_ = data.columns
    data = data.iloc[1:10, 0]
    flags = Flags({var1: pd.Series(UNFLAGGED, index=data.index)})

    tests = [
        ("abs(x)", np.abs(data)),
        ("log(x)", np.log(data)),
        ("exp(x)", np.exp(data)),
    ]

    for test, expected in tests:
        func = _compileGeneric(f"processGeneric(func={test})")
        result = _execGeneric(flags, data, func)
        assert (result == expected).all(axis=None)


def test_bitOps(data):
    var1, var2, *_ = data.columns
    flags = initFlagsLike(data)

    tests = [
        ([var1], "~(x > mean(x))", ~(data[var1] > np.nanmean(data[var1]))),
        ([var1], "(x <= 0) | (0 < x)", (data[var1] <= 0) | (0 < data[var1])),
        ([var1, var2], "(y>= 0) & (0 > x)", (data[var2] >= 0) & (0 > data[var1])),
    ]

    for field, test, expected in tests:
        func = _compileGeneric(f"flagGeneric(func={test})")
        result = _execGeneric(Flags({f: flags[f] for f in field}), data[field], func)
        assert (result == expected).all(axis=None)


def test_variableAssignments(data):

    config = f"""
    varname ; test
    dummy1  ; processGeneric(field=["var1", "var2"], func=x + y)
    dummy2  ; flagGeneric(field=["var1", "var2"], func=x + y > 0)
    """

    fobj = writeIO(config)
    saqc = fromConfig(fobj, data)

    expected_columns = set(data.columns) | {"dummy1", "dummy2"}
    assert set(saqc.data.columns) == expected_columns
    assert set(saqc.flags.columns) == expected_columns


def test_processExistingTarget(data):
    config = f"""
    varname ; test
    var2   ; flagMissing()
    var2   ; processGeneric(func=y - 1, flag=DOUBTFUL)
    """

    fobj = writeIO(config)
    saqc = fromConfig(fobj, data)
    assert (saqc._data["var2"] == data["var2"] - 1).all()
    assert len(saqc._flags.history["var2"]) == 2
    assert saqc._flags.history["var2"].hist[0].isna().all()
    assert (saqc._flags.history["var2"].hist[1] == DOUBTFUL).all()


def test_flagTargetExisting(data):
    config = f"""
    varname ; test
    dummy   ; processGeneric(field="var1", func=x < 1)
    dummy   ; processGeneric(field="var2", func=y >1)
    """

    fobj = writeIO(config)
    saqc = fromConfig(fobj, data)
    assert len(saqc.data["dummy"]) == len(saqc.flags["dummy"])


def test_processTargetExistingFail(data_diff):
    config = f"""
    varname ; test
    dummy   ; processGeneric(field="var1", func=x + 1)
    dummy   ; processGeneric(field="var2", func=y - 1)
    """

    fobj = writeIO(config)
    with pytest.raises(ValueError):
        fromConfig(fobj, data_diff)


def test_flagTargetExistingFail(data_diff):
    config = f"""
    varname ; test
    dummy   ; flagGeneric(field="var1", func=x < 1)
    dummy   ; flagGeneric(field="var2", func=y > 1)
    """

    fobj = writeIO(config)
    with pytest.raises(ValueError):
        fromConfig(fobj, data_diff)


def test_callableArgumentsUnary(data):

    window = 5

    @register(mask=["field"], demask=["field"], squeeze=["field"])
    def testFuncUnary(data, field, flags, func, **kwargs):
        value = data[field].rolling(window=window).apply(func)
        data[field] = value
        return data, initFlagsLike(data)

    var = data.columns[0]

    config = f"""
    varname ; test
    {var}   ; testFuncUnary(func={{0}})
    """

    tests = [
        # ("sum", np.nansum),
        ("std(exp(x))", lambda x: np.std(np.exp(x))),
    ]

    for (name, func) in tests:
        fobj = writeIO(config.format(name))
        result_config = fromConfig(fobj, data).result.data
        result_api = SaQC(data).testFuncUnary(var, func=func).result.data
        expected = data[var].rolling(window=window).apply(func)
        assert (result_config[var].dropna() == expected.dropna()).all(axis=None)
        assert (result_api[var].dropna() == expected.dropna()).all(axis=None)


def test_callableArgumentsBinary(data):
    var1, var2 = data.columns[:2]

    @register(mask=["field"], demask=["field"], squeeze=["field"])
    def testFuncBinary(data, field, flags, func, **kwargs):
        data[field] = func(data[var1], data[var2])
        return data, initFlagsLike(data)

    config = f"""
    varname ; test
    {var1}  ; testFuncBinary(func={{0}})
    """

    tests = [
        ("x + y", lambda x, y: x + y),
        ("y - (x * 2)", lambda y, x: y - (x * 2)),
    ]

    for (name, func) in tests:
        fobj = writeIO(config.format(name))
        result_config = fromConfig(fobj, data).result.data
        result_api = SaQC(data).testFuncBinary(var1, func=func).result.data
        expected = func(data[var1], data[var2])
        assert (result_config[var1].dropna() == expected.dropna()).all(axis=None)
        assert (result_api[var1].dropna() == expected.dropna()).all(axis=None)


def test_isflagged(data):

    var1, var2, *_ = data.columns
    flags = initFlagsLike(data)
    flags[data[var1].index[::2], var1] = BAD

    tests = [
        ([var1], f"isflagged(x)", flags[var1] > UNFLAGGED),
        ([var1], f"isflagged(x)", flags[var1] >= BAD),
        ([var2], f"~isflagged(x)", flags[var2] == UNFLAGGED),
        (
            [var1, var2],
            f"~(x > 999) & (~isflagged(y))",
            ~(data[var1] > 999) & (flags[var2] == UNFLAGGED),
        ),
    ]

    for field, test, expected in tests:
        func = _compileGeneric(f"flagGeneric(func={test}, flag=BAD)")
        result = _execGeneric(Flags({f: flags[f] for f in field}), data[field], func)
        assert (result == expected).all(axis=None)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


# see test/functs/fixtures.py for global fixtures "course_..."
import pytest

import dios

from saqc import BAD, UNFLAGGED
from saqc.core import initFlagsLike
from saqc.funcs.tools import maskTime
from saqc.funcs.residues import calculatePolynomialResidues, calculateRollingResidues

from tests.fixtures import *


@pytest.mark.filterwarnings("ignore: The fit may be poorly conditioned")
@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_2")])
def test_modelling_polyFit_forRegular(dat):
    data, _ = dat(
        freq="10min", periods=30, initial_level=0, final_level=100, out_val=-100
    )
    # add some nice sine distortion
    data = data + 10 * np.sin(np.arange(0, len(data.indexes[0])))
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    result1, _ = calculatePolynomialResidues(data, "data", flags, 11, 2, numba=False)
    result2, _ = calculatePolynomialResidues(data, "data", flags, 11, 2, numba=True)
    assert (result1["data"] - result2["data"]).abs().max() < 10 ** -10
    result3, _ = calculatePolynomialResidues(
        data, "data", flags, "110min", 2, numba=False
    )
    assert result3["data"].equals(result1["data"])
    result4, _ = calculatePolynomialResidues(
        data, "data", flags, 11, 2, numba=True, min_periods=11
    )
    assert (result4["data"] - result2["data"]).abs().max() < 10 ** -10
    data.iloc[13:16] = np.nan
    result5, _ = calculatePolynomialResidues(
        data, "data", flags, 11, 2, numba=True, min_periods=9
    )
    assert result5["data"].iloc[10:19].isna().all()


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_2")])
def test_modelling_rollingMean_forRegular(dat):
    data, _ = dat(
        freq="10min", periods=30, initial_level=0, final_level=100, out_val=-100
    )
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    calculateRollingResidues(
        data,
        "data",
        flags,
        5,
        func=np.mean,
        min_periods=0,
        center=True,
    )
    calculateRollingResidues(
        data,
        "data",
        flags,
        5,
        func=np.mean,
        min_periods=0,
        center=False,
    )


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_1")])
def test_modelling_mask(dat):
    data, _ = dat()
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    field = "data"

    # set flags everywhere to test unflagging
    flags[:, field] = BAD

    common = dict(data=data, field=field, flags=flags, mode="periodic")
    data_seasonal, flags_seasonal = maskTime(
        **common, start="20:00", end="40:00", closed=False
    )
    flagscol = flags_seasonal[field]
    m = (20 <= flagscol.index.minute) & (flagscol.index.minute <= 40)
    assert all(flags_seasonal[field][m] == UNFLAGGED)
    assert all(data_seasonal[field][m].isna())

    data_seasonal, flags_seasonal = maskTime(**common, start="15:00:00", end="02:00:00")
    flagscol = flags_seasonal[field]
    m = (15 <= flagscol.index.hour) & (flagscol.index.hour <= 2)
    assert all(flags_seasonal[field][m] == UNFLAGGED)
    assert all(data_seasonal[field][m].isna())

    data_seasonal, flags_seasonal = maskTime(
        **common, start="03T00:00:00", end="10T00:00:00"
    )
    flagscol = flags_seasonal[field]
    m = (3 <= flagscol.index.hour) & (flagscol.index.hour <= 10)
    assert all(flags_seasonal[field][m] == UNFLAGGED)
    assert all(data_seasonal[field][m].isna())

    mask_ser = pd.Series(False, index=data["data"].index)
    mask_ser[::5] = True
    data["mask_ser"] = mask_ser
    flags = initFlagsLike(data)
    data_masked, flags_masked = maskTime(
        data, "data", flags, mode="mask_field", mask_field="mask_ser"
    )
    m = mask_ser
    assert all(flags_masked[field][m] == UNFLAGGED)
    assert all(data_masked[field][m].isna())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from functools import partial
from pathlib import Path

import click

import numpy as np
import pandas as pd
import pyarrow as pa

from saqc.core.reader import fromConfig
from saqc.core.core import TRANSLATION_SCHEMES


logger = logging.getLogger("SaQC")


def _setupLogging(loglvl):
    logger.setLevel(loglvl)
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def setupIO(nodata):
    reader = {
        ".csv": partial(pd.read_csv, index_col=0, parse_dates=True),
        ".parquet": pd.read_parquet,
    }

    writer = {
        ".csv": partial(pd.DataFrame.to_csv, header=True, index=True, na_rep=nodata),
        ".parquet": lambda df, outfile: pa.parquet.write_table(
            pa.Table.from_pandas(df), outfile
        ),
    }
    return reader, writer


def readData(reader_dict, fname):
    extension = Path(fname).suffix
    reader = reader_dict.get(extension)
    if not reader:
        raise ValueError(
            f"Unsupported file format '{extension}', use one of {tuple(reader.keys())}"
        )
    return reader(fname)


def writeData(writer_dict, df, fname):
    extension = Path(fname).suffix
    writer = writer_dict.get(extension)
    if not writer:
        raise ValueError(
            f"Unsupported file format '{extension}', use one of {tuple(writer.keys())}"
        )
    writer(df, fname)


@click.command()
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    required=True,
    help="path to the configuration file",
)
@click.option(
    "-d",
    "--data",
    type=click.Path(exists=True),
    multiple=True,
    required=True,
    help="path to the data file",
)
@click.option(
    "-o", "--outfile", type=click.Path(exists=False), help="path to the output file"
)
@click.option(
    "--scheme",
    default=None,
    type=click.Choice(tuple(TRANSLATION_SCHEMES.keys())),
    help="the flagging scheme to use",
)
@click.option("--nodata", default=np.nan, help="nodata value")
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
    help="set output verbosity",
)
def main(config, data, scheme, outfile, nodata, log_level):
    # data is always a list of data files

    _setupLogging(log_level)
    reader, writer = setupIO(nodata)

    _data = []
    for dfile in data:
        df = readData(reader, dfile)
        _data.append(df)
    data = _data

    saqc = fromConfig(
        config,
        data=data,
        scheme=TRANSLATION_SCHEMES[scheme or "simple"](),
    )

    data_result, flags_result = saqc.result.data, saqc.result.flags

    if outfile:

        data_result.columns = pd.MultiIndex.from_product(
            [data_result.columns.tolist(), ["data"]]
        )

        if not isinstance(flags_result.columns, pd.MultiIndex):
            flags_result.columns = pd.MultiIndex.from_product(
                [flags_result.columns.tolist(), ["flags"]]
            )

        out = pd.concat([data_result, flags_result], axis=1).sort_index(
            axis=1, level=0, sort_remaining=False
        )

        writeData(writer, out, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import pytest
import numpy as np
import pandas as pd

from saqc.core.history import History, createHistoryFromData
from tests.common import dummyHistory

# see #GH143 combined backtrack
# (adjusted to current implementation)
example1 = (
    # flags
    np.array(
        [
            [0, np.nan, 50, 99, np.nan],
            [0, np.nan, 50, np.nan, 25],
            [0, 99, 99, 99, 25],
            [0, 99, np.nan, np.nan, 25],
        ]
    ),
    # expected from max()
    np.array([99, 25, 25, 25]),
)

# see #GH143
example2 = (
    # flags
    np.array(
        [
            [0, 99, np.nan, 0],
            [0, np.nan, 99, np.nan],
            [0, np.nan, np.nan, np.nan],
            [0, np.nan, np.nan, 0],
        ]
    ),
    # expected from max()
    np.array([0, 99, 0, 0]),
)

data = [
    np.array([[]]),
    np.zeros((1, 1)),
    np.zeros((3, 4)),
    np.ones((3, 4)),
    np.ones((3, 4)) * np.nan,
    np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 2, 3],
            [0, 1, 2, 3],
        ]
    ),
    np.array(
        [
            [0, 0, 0, 0],
            [0, 1, np.nan, 3],
            [0, 1, 2, 3],
        ]
    ),
]


def check_invariants(hist):
    """
    This can be called for **any** FH.
    The assertions must hold in any case.
    """
    # basics
    assert isinstance(hist, History)
    assert isinstance(hist.hist, pd.DataFrame)
    assert isinstance(hist.meta, list)
    assert all(
        [isinstance(dtype, (float, pd.CategoricalDtype)) for dtype in hist.hist.dtypes]
    )
    assert all([isinstance(e, dict) for e in hist.meta])
    assert hist.columns is hist.hist.columns
    assert hist.index is hist.hist.index
    assert len(hist) == len(hist.columns) == len(hist.meta)

    # advanced
    assert hist.columns.equals(pd.Index(range(len(hist))))
    assert isinstance(hist.max(), pd.Series)


def is_equal(hist1: History, hist2: History):
    """
    Check if two FH are (considered) equal, namely have equal 'hist'
    """
    return hist1.hist.equals(hist2.hist)


@pytest.mark.parametrize("data", data + [None])
def test_init(data: np.array):
    # init
    df = pd.DataFrame(data, dtype=float)
    hist = History(df.index)
    check_invariants(hist)


@pytest.mark.parametrize("data", data + [None])
def test_createHistory(data: np.array):
    # init
    df = pd.DataFrame(data, dtype=float)
    meta = [{}] * len(df.columns)
    hist = createHistoryFromData(df, meta)

    check_invariants(hist)

    # shape would fail
    if data is not None:
        assert len(hist.index) == data.shape[0]
        assert len(hist.columns) == data.shape[1]


@pytest.mark.parametrize("data", data + [None])
def test_copy(data):
    # init
    df = pd.DataFrame(data, dtype=float)
    hist = History(df.index)
    for _, s in df.items():
        hist.append(s)
    shallow = hist.copy(deep=False)
    deep = hist.copy(deep=True)

    # checks
    for copy in [deep, shallow]:
        check_invariants(copy)
        assert copy is not hist
        assert copy.hist is not hist.hist
        assert copy.meta is not hist.meta
        assert is_equal(copy, hist)

    assert deep is not shallow
    assert is_equal(deep, shallow)

    # underling pandas data was only copied with deep=True
    assert shallow.hist.index is hist.hist.index
    assert deep.hist.index is not hist.hist.index


@pytest.mark.parametrize("copy", [True, False])
@pytest.mark.parametrize("data", data + [None])
def test_reindex_trivial_cases(data, copy):
    df = pd.DataFrame(data, dtype=float)
    orig = dummyHistory(hist=df)

    # checks
    for index in [df.index, pd.Index([])]:
        ref = orig.copy()
        hist = ref.reindex(index, copy=copy)
        if copy:
            assert hist is not ref
        else:
            assert hist is ref
        check_invariants(hist)


@pytest.mark.parametrize("copy", [True, False])
@pytest.mark.parametrize("data", data + [None])
def test_reindex_missing_indicees(data, copy):
    df = pd.DataFrame(data, dtype=float)
    orig = dummyHistory(hist=df)
    index = df.index[1:-1]
    hist = orig.reindex(index, copy=copy)
    if copy:
        assert hist is not orig
    else:
        assert hist is orig
    check_invariants(hist)


@pytest.mark.parametrize("copy", [True, False])
@pytest.mark.parametrize("data", data + [None])
def test_reindex_extra_indicees(data, copy):
    df = pd.DataFrame(data, dtype=float)
    orig = dummyHistory(hist=df)
    index = df.index.append(pd.Index(range(len(df.index), len(df.index) + 5)))
    hist = orig.reindex(index, copy=copy)
    if copy:
        assert hist is not orig
    else:
        assert hist is orig
    check_invariants(hist)


@pytest.mark.parametrize(
    "s, meta",
    [
        (pd.Series(0, index=range(6), dtype=float), None),
        (pd.Series(0, index=range(6), dtype=float), {}),
        (pd.Series(1, index=range(6), dtype=float), {"foo": "bar"}),
    ],
)
def test_append_with_meta(s, meta):
    hist = History(s.index)
    hist.append(s, meta=meta)
    check_invariants(hist)

    if meta is None:
        meta = {}

    assert hist.meta[0] is not meta
    assert hist.meta == [meta]

    hist.append(s, meta=meta)
    check_invariants(hist)
    assert hist.meta == [meta, meta]


@pytest.fixture(scope="module")
def __hist():
    # this FH is filled by
    #  - test_append
    #  - test_append_force
    return History(index=pd.Index(range(6)))


# this test append more rows to the resulting
# FH from the former test
@pytest.mark.parametrize(
    "s, max_val",
    [
        (pd.Series(0, index=range(6), dtype=float), 0),
        (pd.Series(1, index=range(6), dtype=float), 1),
        (pd.Series(np.nan, index=range(6), dtype=float), 1),
        (pd.Series(0, index=range(6), dtype=float), 0),
    ],
)
def test_append_force(__hist, s, max_val):
    hist = __hist
    hist.append(s)
    check_invariants(hist)
    assert all(hist.max() == max_val)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


# see test/functs/fixtures.py for global fixtures "course_..."

import dios

from saqc.constants import *
from saqc.core import initFlagsLike
from saqc.funcs.transformation import transform
from saqc.funcs.drift import correctOffset
from saqc.funcs.interpolation import (
    interpolateByRolling,
    interpolateInvalid,
    interpolateIndex,
)
from saqc.funcs.resampling import resample
from saqc.lib.ts_operators import linearInterpolation, polynomialInterpolation

from tests.fixtures import *


def test_rollingInterpolateMissing(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5, 6])
    field = data.columns[0]
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    dataInt, *_ = interpolateByRolling(
        data.copy(),
        field,
        flags.copy(),
        3,
        func=np.median,
        center=True,
        min_periods=0,
        interpol_flag=UNFLAGGED,
    )
    assert dataInt[field][characteristics["missing"]].notna().all()
    dataInt, *_ = interpolateByRolling(
        data.copy(),
        field,
        flags.copy(),
        3,
        func=np.nanmean,
        center=False,
        min_periods=3,
        interpol_flag=UNFLAGGED,
    )
    assert dataInt[field][characteristics["missing"]].isna().all()


def test_interpolateMissing(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5])
    field = data.columns[0]
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    dataLin, *_ = interpolateInvalid(data, field, flags, method="linear")
    dataPoly, *_ = interpolateInvalid(data, field, flags, method="polynomial")
    assert dataLin[field][characteristics["missing"]].notna().all()
    assert dataPoly[field][characteristics["missing"]].notna().all()
    data, characteristics = course_5(periods=10, nan_slice=[5, 6, 7])
    dataLin1, *_ = interpolateInvalid(
        data.copy(), field, flags, method="linear", limit=2
    )
    dataLin2, *_ = interpolateInvalid(
        data.copy(), field, flags, method="linear", limit=3
    )
    dataLin3, *_ = interpolateInvalid(
        data.copy(), field, flags, method="linear", limit=4
    )
    assert dataLin1[field][characteristics["missing"]].isna().all()
    assert dataLin2[field][characteristics["missing"]].isna().all()
    assert dataLin3[field][characteristics["missing"]].notna().all()


def test_transform(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5, 6])
    field = data.columns[0]
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    data1, *_ = transform(data, field, flags, func=linearInterpolation)
    assert data1[field][characteristics["missing"]].isna().all()
    data1, *_ = transform(
        data, field, flags, func=lambda x: linearInterpolation(x, inter_limit=3)
    )
    assert data1[field][characteristics["missing"]].notna().all()
    data1, *_ = transform(
        data,
        field,
        flags,
        func=lambda x: polynomialInterpolation(x, inter_limit=3, inter_order=3),
    )
    assert data1[field][characteristics["missing"]].notna().all()


def test_resample(course_5):
    data, characteristics = course_5(
        freq="1min", periods=30, nan_slice=[1, 11, 12, 22, 24, 26]
    )
    field = data.columns[0]
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    data1, *_ = resample(
        data,
        field,
        flags,
        "10min",
        np.mean,
        maxna=2,
        maxna_group=1,
    )
    assert ~np.isnan(data1[field].iloc[0])
    assert np.isnan(data1[field].iloc[1])
    assert np.isnan(data1[field].iloc[2])


def test_interpolateGrid(course_5, course_3):
    data, _ = course_5()
    data_grid, characteristics = course_3()
    data["grid"] = data_grid.to_df()
    # data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    dataInt, *_ = interpolateIndex(
        data, "data", flags, "1h", "time", grid_field="grid", limit=10
    )


def test_offsetCorrecture():
    data = pd.Series(0, index=pd.date_range("2000", freq="1d", periods=100), name="dat")
    data.iloc[30:40] = -100
    data.iloc[70:80] = 100
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    data, _ = correctOffset(data, "dat", flags, 40, 20, "3d", 1)
    assert (data == 0).all()[0]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import json
from typing import Any
from functools import reduce

import numpy as np
import pandas as pd

from saqc.core.history import History
from saqc.core.flags import (
    Flags,
    UNFLAGGED,
    GOOD,
    DOUBTFUL,
    BAD,
)
from saqc.core.translation.basescheme import (
    BackwardMap,
    TranslationScheme,
    ForwardMap,
)


_QUALITY_CAUSES = [
    "",
    "BATTERY_LOW",
    "BELOW_MINIMUM",
    "ABOVE_MAXIMUM",
    "BELOW_OR_ABOVE_MIN_MAX",
    "ISOLATED_SPIKE",
    "DEFECTIVE_SENSOR",
    "LEFT_CENSORED_DATA",
    "RIGHT_CENSORED_DATA",
    "OTHER",
    "AUTOFLAGGED",
]

_QUALITY_LABELS = [
    "quality_flag",
    "quality_cause",
    "quality_comment",
]


class DmpScheme(TranslationScheme):

    """
    Implements the translation from and to the flagging scheme implemented in
    the UFZ - Datamanagementportal
    """

    ARGUMENTS = {"comment": "", "cause": "OTHER"}

    _FORWARD: ForwardMap = {
        "NIL": UNFLAGGED,
        "OK": GOOD,
        "DOUBTFUL": DOUBTFUL,
        "BAD": BAD,
    }

    _BACKWARD: BackwardMap = {
        UNFLAGGED: "NIL",
        np.nan: "NIL",
        GOOD: "OK",
        DOUBTFUL: "DOUBTFUL",
        BAD: "BAD",
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)

    def forward(self, df: pd.DataFrame) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        df : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """

        self.validityCheck(df)

        data = {}

        for field in df.columns.get_level_values(0):

            field_flags = df[field]
            field_history = History(field_flags.index)

            for (flag, cause, comment), values in field_flags.groupby(_QUALITY_LABELS):
                try:
                    comment = json.loads(comment)
                except json.decoder.JSONDecodeError:
                    comment = {"test": "unknown", "comment": ""}

                histcol = pd.Series(np.nan, index=field_flags.index)
                histcol.loc[values.index] = self(flag)

                meta = {
                    "func": comment["test"],
                    "kwargs": {"comment": comment["comment"], "cause": cause},
                }
                field_history.append(histcol, meta=meta)

            data[str(field)] = field_history

        return Flags(data)

    def backward(
        self, flags: Flags, attrs: dict | None = None, **kwargs
    ) -> pd.DataFrame:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : The external flags to translate

        attrs : dict or None, default None
            global meta information of saqc-object

        Returns
        -------
        translated flags
        """
        tflags = super().backward(flags, raw=True, attrs=attrs)

        out = pd.DataFrame(
            index=reduce(lambda x, y: x.union(y), tflags.indexes).sort_values(),
            columns=pd.MultiIndex.from_product([flags.columns, _QUALITY_LABELS]),
        )

        for field in tflags.columns:

            df = pd.DataFrame(
                {
                    "quality_flag": tflags[field],
                    "quality_cause": "",
                    "quality_comment": "",
                }
            )

            history = flags.history[field]
            for col in history.columns:

                valid = (history.hist[col] != UNFLAGGED) & history.hist[col].notna()

                # extract from meta
                meta = history.meta[col]
                keywords = meta.get("kwargs", {})
                comment = json.dumps(
                    {
                        "test": meta.get("func", "unknown"),
                        "comment": keywords.get("comment", self.ARGUMENTS["comment"]),
                    }
                )
                cause = keywords.get("cause", self.ARGUMENTS["cause"])
                df.loc[valid, "quality_comment"] = comment
                df.loc[valid, "quality_cause"] = cause

            out[field] = df.reindex(out.index)

        self.validityCheck(out)
        return out

    @classmethod
    def validityCheck(cls, df: pd.DataFrame) -> None:
        """
        Check wether the given causes and comments are valid.

        Parameters
        ----------
        df : external flags
        """

        cols = df.columns
        if not isinstance(cols, pd.MultiIndex):
            raise TypeError("DMP-Flags need multi-index columns")

        if not cols.get_level_values(1).isin(_QUALITY_LABELS).all(axis=None):
            raise TypeError(
                f"DMP-Flags expect the labels {list(_QUALITY_LABELS)} in the secondary level"
            )

        for field in df.columns.get_level_values(0):

            # we might have NaN injected by DictOfSeries -> DataFrame conversions
            field_df = df[field].dropna(how="all", axis="index")
            flags = field_df["quality_flag"]
            causes = field_df["quality_cause"]
            comments = field_df["quality_comment"]

            if not flags.isin(cls._FORWARD.keys()).all(axis=None):
                raise ValueError(
                    f"invalid quality flag(s) found, only the following values are supported: {set(cls._FORWARD.keys())}"
                )

            if not causes.isin(_QUALITY_CAUSES).all(axis=None):
                raise ValueError(
                    f"invalid quality cause(s) found, only the following values are supported: {_QUALITY_CAUSES}"
                )

            if (~flags.isin(("OK", "NIL")) & (causes == "")).any(axis=None):
                raise ValueError(
                    "quality flags other than 'OK and 'NIL' need a non-empty quality cause"
                )

            if ((causes == "OTHER") & (comments == "")).any(None):
                raise ValueError(
                    "quality cause 'OTHER' needs a non-empty quality comment"
                )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import functools
from typing import Optional, Tuple, Sequence, Callable, Optional
from typing_extensions import Literal

import numpy as np
import pandas as pd

from scipy import stats
from scipy.optimize import curve_fit
from scipy.spatial.distance import pdist

from dios import DictOfSeries
from saqc.constants import BAD
from saqc.core.register import register, flagging
from saqc.core import Flags
from saqc.funcs.changepoints import _assignChangePointCluster
from saqc.funcs.tools import dropField, copyField

from saqc.lib.tools import detectDeviants, toSequence, filterKwargs
from saqc.lib.ts_operators import linearDriftModel, expDriftModel
from saqc.lib.types import CurveFitter
from saqc.lib.ts_operators import linearDriftModel, expDriftModel


LinkageString = Literal[
    "single", "complete", "average", "weighted", "centroid", "median", "ward"
]

MODELDICT = {"linear": linearDriftModel, "exponential": expDriftModel}


@flagging()
def flagDriftFromNorm(
    data: DictOfSeries,
    field: Sequence[str],
    flags: Flags,
    freq: str,
    spread: float,
    frac: float = 0.5,
    metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
        np.array([x, y]), metric="cityblock"
    )
    / len(x),
    method: LinkageString = "single",
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags value courses that significantly deviate from a group of normal value courses.

    "Normality" is determined in terms of a maximum spreading distance, that members of a normal group must not exceed.
    In addition, only a group is considered "normal" if it contains more then `frac` percent of the
    variables in "field".

    See the Notes section for a more detailed presentation of the algorithm

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : list of str
        List of fieldnames in data, determining which variables are to be included into the flagging process.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    freq : str
        An offset string, determining the size of the seperate datachunks that the algorihm is to be piecewise
        applied on.
    spread : float
        A parameter limiting the maximum "spread" of the timeseries, allowed in the "normal" group. See Notes section
        for more details.
    frac : float, default 0.5
        Has to be in [0,1]. Determines the minimum percentage of variables, the "normal" group has to comprise to be the
        normal group actually. The higher that value, the more stable the algorithm will be with respect to false
        positives. Also, nobody knows what happens, if this value is below 0.5.
    metric : Callable[[numpy.array, numpy.array], float]
        A distance function. It should be a function of 2 1-dimensional arrays and return a float scalar value.
        This value is interpreted as the distance of the two input arrays. The default is the averaged manhatten metric.
        See the Notes section to get an idea of why this could be a good choice.
    method : {"single", "complete", "average", "weighted", "centroid", "median", "ward"}, default "single"
        The linkage method used for hierarchical (agglomerative) clustering of the timeseries.
        See the Notes section for more details.
        The keyword gets passed on to scipy.hierarchy.linkage. See its documentation to learn more about the different
        keywords (References [1]).
        See wikipedia for an introduction to hierarchical clustering (References [2]).
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the input flags.

    Notes
    -----
    following steps are performed for every data "segment" of length `freq` in order to find the
    "abnormal" data:

    1. Calculate the distances :math:`d(x_i,x_j)` for all :math:`x_i` in parameter `field`. (with :math:`d`
       denoting the distance function
       passed to the parameter `metric`.
    2. Calculate a dendogram with a hierarchical linkage algorithm, specified by the parameter `method`.
    3. Flatten the dendogram at the level, the agglomeration costs exceed the value given by the parameter `spread`
    4. check if there is a cluster containing more than `frac` percentage of the variables in field.

        1. if yes: flag all the variables that are not in that cluster (inside the segment)
        2. if no: flag nothing

    The main parameter giving control over the algorithms behavior is the `spread` parameter, that determines
    the maximum spread of a normal group by limiting the costs, a cluster agglomeration must not exceed in every
    linkage step.
    For singleton clusters, that costs just equal half the distance, the timeseries in the clusters, have to
    each other. So, no timeseries can be clustered together, that are more then
    2*`spread` distanted from each other.
    When timeseries get clustered together, this new clusters distance to all the other timeseries/clusters is
    calculated according to the linkage method specified by `method`. By default, it is the minimum distance,
    the members of the clusters have to each other.
    Having that in mind, it is advisable to choose a distance function, that can be well interpreted in the units
    dimension of the measurement and where the interpretation is invariant over the length of the timeseries.
    That is, why, the "averaged manhatten metric" is set as the metric default, since it corresponds to the
    averaged value distance, two timeseries have (as opposed by euclidean, for example).

    References
    ----------
    Documentation of the underlying hierarchical clustering algorithm:
        [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
    Introduction to Hierarchical clustering:
        [2] https://en.wikipedia.org/wiki/Hierarchical_clustering
    """
    fields = toSequence(field)
    data_to_flag = data[fields].to_df()
    data_to_flag.dropna(inplace=True)

    segments = data_to_flag.groupby(pd.Grouper(freq=freq))
    for segment in segments:

        if segment[1].shape[0] <= 1:
            continue

        drifters = detectDeviants(segment[1], metric, spread, frac, method, "variables")

        for var in drifters:
            flags[segment[1].index, fields[var]] = flag

    return data, flags


# default multivariate flagging
@register(
    mask=["field", "reference"],
    demask=["field", "reference"],
    squeeze=["field", "reference"],  # reference is written !
    multivariate=True,
    handles_target=True,
)
def flagDriftFromReference(
    data: DictOfSeries,
    field: Sequence[str],
    flags: Flags,
    reference: str,
    freq: str,
    thresh: float,
    metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
        np.array([x, y]), metric="cityblock"
    )
    / len(x),
    target=None,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags value courses that deviate from a reference course by a margin exceeding a certain threshold.

    The deviation is measured by the distance function passed to parameter metric.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : list of str
        List of fieldnames in data, determining wich variables are to be included into the flagging process.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    reference : str
        The reference variable, the deviation from wich determines the flagging.
    freq : str
        An offset string, determining the size of the seperate datachunks that the algorihm is to be piecewise
        applied on.
    thresh : float
        The threshod by wich normal variables can deviate from the reference variable at max.
    metric : Callable[(numpyp.array, numpy-array), float]
        A distance function. It should be a function of 2 1-dimensional arrays and return a float scalar value.
        This value is interpreted as the distance of the two input arrays. The default is the averaged manhatten metric.
        See the Notes section to get an idea of why this could be a good choice.
    target : None
        ignored.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the input flags.

    Notes
    -----
    it is advisable to choose a distance function, that can be well interpreted in the units
    dimension of the measurement and where the interpretation is invariant over the length of the timeseries.
    That is, why, the "averaged manhatten metric" is set as the metric default, since it corresponds to the
    averaged value distance, two timeseries have (as opposed by euclidean, for example).
    """
    if target:
        raise NotImplementedError("target is not implemented for this function")

    fields = toSequence(field)
    data_to_flag = data[fields].to_df().dropna()

    fields = list(fields)
    if reference not in fields:
        fields.append(reference)

    segments = data_to_flag.groupby(pd.Grouper(freq=freq))
    for segment in segments:

        if segment[1].shape[0] <= 1:
            continue

        for i in range(len(fields)):
            dist = metric(
                segment[1].iloc[:, i].values, segment[1].loc[:, reference].values
            )

            if dist > thresh:
                flags[segment[1].index, fields[i]] = flag

    return data, flags


# default multivariate flagging
@register(
    mask=["set_1", "set_2"],
    demask=["set_1", "set_2"],
    squeeze=["set_1", "set_2"],
    multivariate=True,
    handles_target=True,
)
def flagDriftFromScaledNorm(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    set_1: Sequence[str],
    set_2: Sequence[str],
    freq: str,
    spread: float,
    frac: float = 0.5,
    metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
        np.array([x, y]), metric="cityblock"
    )
    / len(x),
    method: LinkageString = "single",
    target: str = None,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function linearly rescales one set of variables to another set of variables
    with a different scale and then flags value courses that significantly deviate
    from a group of normal value courses.

    The two sets of variables can be linearly scaled one to another and hence the
    scaling transformation is performed via linear regression: A linear regression is
    performed on each pair of variables giving a slope and an intercept. The
    transformation is then calculated a the median of all the calculated slopes and
    intercepts.

    Once the transformation is performed, the function flags those values,
    that deviate from a group of normal values. "Normality" is determined in terms of
    a maximum spreading distance, that members of a normal group must not exceed. In
    addition, only a group is considered "normal" if it contains more then `frac`
    percent of the variables in "fields".

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        A dummy parameter.

    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.

    set_1 : str
        The first set of fieldnames in data to be included into the flagging process.

    set_2 : str
        The second set of fieldnames in data to be included into the flagging process.

    freq : str
        An offset string, determining the size of the seperate datachunks that the
        algorihm is to be piecewise applied on.

    spread : float
        A parameter limiting the maximum "spread" of the timeseries, allowed in the
        "normal" group. See Notes section for more details.

    frac : float, default 0.5
        Has to be in [0,1]. Determines the minimum percentage of variables,
        the "normal" group has to comprise to be the normal group actually. The
        higher that value, the more stable the algorithm will be with respect to
        false positives. Also, nobody knows what happens, if this value is below 0.5.

    metric : Callable[(numpyp.array, numpy-array), float]
        A distance function. It should be a function of 2 1-dimensional arrays and
        return a float scalar value. This value is interpreted as the distance of the
        two input arrays. The default is the averaged manhatten metric. See the Notes
        section to get an idea of why this could be a good choice.

    method : str, default "single"
        The linkage method used for hierarchical (agglomerative) clustering of the
        timeseries. See the Notes section for more details. The keyword gets passed
        on to scipy.hierarchy.linkage. See its documentation to learn more about the
        different keywords (References [1]). See wikipedia for an introduction to
        hierarchical clustering (References [2]).

    target : None
        ignored

    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the input flags.

    References
    ----------
    Documentation of the underlying hierarchical clustering algorithm:
        [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
    Introduction to Hierarchical clustering:
        [2] https://en.wikipedia.org/wiki/Hierarchical_clustering
    """
    if target:
        raise NotImplementedError("target is not implemented for this function")

    fields = list(set_1) + list(set_2)
    data_to_flag = data[fields].to_df().dropna()

    convert_slope = []
    convert_intercept = []

    for field1 in set_1:
        for field2 in set_2:
            slope, intercept, *_ = stats.linregress(
                data_to_flag[field1], data_to_flag[field2]
            )
            convert_slope.append(slope)
            convert_intercept.append(intercept)

    factor_slope = np.median(convert_slope)
    factor_intercept = np.median(convert_intercept)

    dat = DictOfSeries()
    for field1 in set_1:
        dat[field1] = factor_intercept + factor_slope * data_to_flag[field1]
    for field2 in set_2:
        dat[field2] = data_to_flag[field2]

    dat_to_flag = dat[fields].to_df()

    segments = dat_to_flag.groupby(pd.Grouper(freq=freq))
    for segment in segments:

        if segment[1].shape[0] <= 1:
            continue

        drifters = detectDeviants(segment[1], metric, spread, frac, method, "variables")

        for var in drifters:
            flags[segment[1].index, fields[var]] = flag

    return data, flags


@register(mask=["field"], demask=[], squeeze=[])
def correctDrift(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    maintenance_field: str,
    model: Callable[..., float] | Literal["linear", "exponential"],
    cal_range: int = 5,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function corrects drifting behavior.

    See the Notes section for an overview over the correction algorithm.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to correct.
    flags : saqc.Flags
        Container to store quality flags to data.
    maintenance_field : str
        The fieldname of the datacolumn holding the support-points information.
        The maint data is to expected to have following form:
        The series' timestamp itself represents the beginning of a
        maintenance event, wheras the values represent the endings of the maintenance intervals.
    model : Callable or {'exponential', 'linear'}
        A modelfunction describing the drift behavior, that is to be corrected.
        Either use built-in exponential or linear drift model by passing a string, or pass a custom callable.
        The model function must always contain the keyword parameters 'origin' and 'target'.
        The starting parameter must always be the parameter, by wich the data is passed to the model.
        After the data parameter, there can occure an arbitrary number of model calibration arguments in
        the signature.
        See the Notes section for an extensive description.
    cal_range : int, default 5
        The number of values the mean is computed over, for obtaining the value level directly after and
        directly before maintenance event. This values are needed for shift calibration. (see above description)

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data

    Notes
    -----
    It is assumed, that between support points, there is a drift effect shifting the meassurements in a way, that
    can be described, by a model function ``M(t, *p, origin, target)``. (With ``0<=t<=1``, ``p`` being a parameter
    set, and ``origin``, ``target`` being floats).

    Note, that its possible for the model to have no free parameters p at all. (linear drift mainly)

    The drift model, directly after the last support point (``t=0``),
    should evaluate to the origin - calibration level (``origin``), and directly before the next support point
    (``t=1``), it should evaluate to the target calibration level (``target``).::

        M(0, *p, origin, target) = origin
        M(1, *p, origin, target) = target

    The model is than fitted to any data chunk in between support points, by optimizing the parameters ``p*``, and
    thus, obtaining optimal parameterset ``P*``.

    The new values at t are computed via:::

        new_vals(t) = old_vals(t) + M(t, *P, origin, target) - M_drift(t, *P, origin, new_target)

    Wheras ``new_target`` represents the value level immediately after the next support point.

    Examples
    --------
    Some examples of meaningful driftmodels.

    Linear drift modell (no free parameters).

    >>> M = lambda t, origin, target: origin + t * target

    exponential drift model (exponential raise!)

    >>> expFunc = lambda t, a, b, c: a + b * (np.exp(c * x) - 1)
    >>> M = lambda t, p, origin, target: expFunc(t, (target - origin) / (np.exp(abs(c)) - 1), abs(c))

    Exponential and linear driftmodels are part of the ``ts_operators`` library, under the names
    ``expDriftModel`` and ``linearDriftModel``.

    """
    # extract model func:
    if isinstance(model, str):
        if model not in MODELDICT:
            raise ValueError(
                f"invalid model '{model}', choose one of '{MODELDICT.keys()}'"
            )
        model = MODELDICT[model]

    # 1: extract fit intervals:
    if data[maintenance_field].empty:
        return data, flags

    to_correct = data[field].copy()
    maint_data = data[maintenance_field].copy()

    to_correct_clean = to_correct.dropna()
    d = {"drift_group": np.nan, to_correct.name: to_correct_clean.values}
    drift_frame = pd.DataFrame(d, index=to_correct_clean.index)

    # group the drift frame
    for k in range(0, maint_data.shape[0] - 1):
        # assign group numbers for the timespans in between one maintenance ending and the beginning of the next
        # maintenance time itself remains np.nan assigned
        drift_frame.loc[
            maint_data.values[k] : pd.Timestamp(maint_data.index[k + 1]), "drift_group"
        ] = k

    # define target values for correction
    drift_grouper = drift_frame.groupby("drift_group")
    shift_targets = drift_grouper.aggregate(lambda x: x[:cal_range].mean()).shift(-1)

    for k, group in drift_grouper:
        data_series = group[to_correct.name]
        data_fit, data_shiftTarget = _driftFit(
            data_series, shift_targets.loc[k, :][0], cal_range, model
        )
        data_fit = pd.Series(data_fit, index=group.index)
        data_shiftTarget = pd.Series(data_shiftTarget, index=group.index)
        data_shiftVektor = data_shiftTarget - data_fit
        shiftedData = data_series + data_shiftVektor
        to_correct[shiftedData.index] = shiftedData

    data[field] = to_correct

    return data, flags


@register(mask=["field", "cluster_field"], demask=["cluster_field"], squeeze=[])
def correctRegimeAnomaly(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    cluster_field: str,
    model: CurveFitter,
    tolerance: Optional[str] = None,
    epoch: bool = False,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function fits the passed model to the different regimes in data[field] and tries to correct
    those values, that have assigned a negative label by data[cluster_field].

    Currently, the only correction mode supported is the "parameter propagation."

    This means, any regime :math:`z`, labeled negatively and being modeled by the parameters p, gets corrected via:

    :math:`z_{correct} = z + (m(p^*) - m(p))`,

    where :math:`p^*` denotes the parameter set belonging to the fit of the nearest not-negatively labeled cluster.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to correct.
    flags : saqc.Flags
        Container to store flags of the data.
    cluster_field : str
        A string denoting the field in data, holding the cluster label for the data you want to correct.
    model : Callable
        The model function to be fitted to the regimes.
        It must be a function of the form :math:`f(x, *p)`, where :math:`x` is the ``numpy.array`` holding the
        independent variables and :math:`p` are the model parameters that are to be obtained by fitting.
        Depending on the `x_date` parameter, independent variable x will either be the timestamps
        of every regime transformed to seconds from epoch, or it will be just seconds, counting the regimes length.
    tolerance : {None, str}, default None:
        If an offset string is passed, a data chunk of length `offset` right at the
        start and right at the end is ignored when fitting the model. This is to account for the
        unreliability of data near the changepoints of regimes.
    epoch : bool, default False
        If True, use "seconds from epoch" as x input to the model func, instead of "seconds from regime start".

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    cluster_ser = data[cluster_field]
    unique_successive = pd.unique(cluster_ser.values)
    data_ser = data[field]
    regimes = data_ser.groupby(cluster_ser)
    para_dict = {}
    x_dict = {}
    x_mask = {}
    if tolerance is not None:
        # get seconds
        tolerance = pd.Timedelta(tolerance).total_seconds()
    for label, regime in regimes:
        if epoch is False:
            # get seconds data:
            xdata = (regime.index - regime.index[0]).to_numpy(dtype=float) * 10 ** (-9)
        else:
            # get seconds from epoch data
            xdata = regime.index.to_numpy(dtype=float) * 10 ** (-9)
        ydata = regime.values
        valid_mask = ~np.isnan(ydata)
        if tolerance is not None:
            valid_mask &= xdata > xdata[0] + tolerance
            valid_mask &= xdata < xdata[-1] - tolerance
        try:
            p, *_ = curve_fit(model, xdata[valid_mask], ydata[valid_mask])
        except (RuntimeError, ValueError):
            p = np.array([np.nan])
        para_dict[label] = p
        x_dict[label] = xdata
        x_mask[label] = valid_mask

    first_normal = unique_successive > 0
    first_valid = np.array(
        [
            ~pd.isna(para_dict[unique_successive[i]]).any()
            for i in range(0, unique_successive.shape[0])
        ]
    )
    first_valid = np.where(first_normal & first_valid)[0][0]
    last_valid = 1

    for k in range(0, unique_successive.shape[0]):
        if unique_successive[k] < 0 & (
            not pd.isna(para_dict[unique_successive[k]]).any()
        ):
            ydata = data_ser[regimes.groups[unique_successive[k]]].values
            xdata = x_dict[unique_successive[k]]
            ypara = para_dict[unique_successive[k]]
            if k > 0:
                target_para = para_dict[unique_successive[k - last_valid]]
            else:
                # first regime has no "last valid" to its left, so we use first valid to the right:
                target_para = para_dict[unique_successive[k + first_valid]]
            y_shifted = ydata + (model(xdata, *target_para) - model(xdata, *ypara))
            data_ser[regimes.groups[unique_successive[k]]] = y_shifted
            if k > 0:
                last_valid += 1
        elif pd.isna(para_dict[unique_successive[k]]).any() & (k > 0):
            last_valid += 1
        else:
            last_valid = 1

    data[field] = data_ser
    return data, flags


@register(mask=["field"], demask=[], squeeze=[])
def correctOffset(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    max_jump: float,
    spread: float,
    window: str,
    min_periods: int,
    tolerance: Optional[str] = None,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to correct.
    flags : saqc.Flags
        Container to store flags of the data.
    max_jump : float
        when searching for changepoints in mean - this is the threshold a mean difference in the
        sliding window search must exceed to trigger changepoint detection.
    spread : float
        threshold denoting the maximum, regimes are allowed to abolutely differ in their means
        to form the "normal group" of values.
    window : str
        Size of the adjacent windows that are used to search for the mean changepoints.
    min_periods : int
        Minimum number of periods a search window has to contain, for the result of the changepoint
        detection to be considered valid.
    tolerance : {None, str}, default None:
        If an offset string is passed, a data chunk of length `offset` right from the
        start and right before the end of any regime is ignored when calculating a regimes mean for data correcture.
        This is to account for the unrelyability of data near the changepoints of regimes.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    data, flags = copyField(data, field, flags, field + "_CPcluster")
    data, flags = _assignChangePointCluster(
        data,
        field + "_CPcluster",
        flags,
        lambda x, y: np.abs(np.mean(x) - np.mean(y)),
        lambda x, y: max_jump,
        window=window,
        min_periods=min_periods,
    )
    data, flags = _assignRegimeAnomaly(data, field, flags, field + "_CPcluster", spread)
    data, flags = correctRegimeAnomaly(
        data,
        field,
        flags,
        field + "_CPcluster",
        lambda x, p1: np.array([p1] * x.shape[0]),
        tolerance=tolerance,
    )
    data, flags = dropField(data, field + "_CPcluster", flags)

    return data, flags


def _driftFit(x, shift_target, cal_mean, driftModel):
    x_index = x.index - x.index[0]
    x_data = x_index.total_seconds().values
    x_data = x_data / x_data[-1]
    y_data = x.values
    origin_mean = np.mean(y_data[:cal_mean])
    target_mean = np.mean(y_data[-cal_mean:])

    dataFitFunc = functools.partial(driftModel, origin=origin_mean, target=target_mean)
    # if drift model has free parameters:
    try:
        # try fitting free parameters
        fit_paras, *_ = curve_fit(dataFitFunc, x_data, y_data)
        data_fit = dataFitFunc(x_data, *fit_paras)
        data_shift = driftModel(
            x_data, *fit_paras, origin=origin_mean, target=shift_target
        )
    except RuntimeError:
        # if fit fails -> make no correction
        data_fit = np.array([0] * len(x_data))
        data_shift = np.array([0] * len(x_data))
    # when there are no free parameters in the model:
    except ValueError:
        data_fit = dataFitFunc(x_data)
        data_shift = driftModel(x_data, origin=origin_mean, target=shift_target)

    return data_fit, data_shift


@flagging()
def flagRegimeAnomaly(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    cluster_field: str,
    spread: float,
    method: LinkageString = "single",
    metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: np.abs(
        np.nanmean(x) - np.nanmean(y)
    ),
    frac: float = 0.5,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flags anomalous regimes regarding to modelling regimes of field.

    "Normality" is determined in terms of a maximum spreading distance,
    regimes must not exceed in respect to a certain metric and linkage method.

    In addition, only a range of regimes is considered "normal", if it models
    more then `frac` percentage of the valid samples in "field".

    Note, that you must detect the regime changepoints prior to calling this function.

    Note, that it is possible to perform hypothesis tests for regime equality
    by passing the metric a function for p-value calculation and selecting linkage
    method "complete".

    Parameters
    ----------
    data : dios.DictOfSeries
        Data to process
    field : str
        Name of the column to process
    flags : saqc.Flags
        Container to store flags of the data.
    cluster_field : str
        Column in data, holding the cluster labels for the samples in field.
        (has to be indexed equal to field)
    spread : float
        A threshold denoting the value level, up to wich clusters a agglomerated.
    method : {"single", "complete", "average", "weighted", "centroid", "median", "ward"}, default "single"
        The linkage method for hierarchical (agglomerative) clustering of the variables.
    metric : Callable, default lambda x,y: np.abs(np.nanmean(x) - np.nanmean(y))
        A metric function for calculating the dissimilarity between 2 regimes.
        Defaults to the difference in mean.
    frac : float
        Has to be in [0,1]. Determines the minimum percentage of samples,
        the "normal" group has to comprise to be the normal group actually.
    flag : float, default BAD
        flag to set.

    Returns
    -------

    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The flags object, holding flags and additional informations related to `data`.
        Flags values may have changed, relatively to the flags input.
    """
    reserverd = ["set_cluster", "set_flags"]
    kwargs = filterKwargs(kwargs, reserverd)
    return _assignRegimeAnomaly(
        data=data,
        field=field,
        flags=flags,
        cluster_field=cluster_field,
        spread=spread,
        method=method,
        metric=metric,
        frac=frac,
        flag=flag,
        **kwargs,
        set_cluster=False,
        set_flags=True,
    )


@register(mask=["field", "cluster_field"], demask=["cluster_field"], squeeze=[])
def assignRegimeAnomaly(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    cluster_field: str,
    spread: float,
    method: LinkageString = "single",
    metric: Callable[[np.array, np.array], float] = lambda x, y: np.abs(
        np.nanmean(x) - np.nanmean(y)
    ),
    frac: float = 0.5,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    A function to detect values belonging to an anomalous regime regarding modelling
    regimes of field.

    The function changes the value of the regime cluster labels to be negative.
    "Normality" is determined in terms of a maximum spreading distance, regimes must
    not exceed in respect to a certain metric and linkage method. In addition,
    only a range of regimes is considered "normal", if it models more then `frac`
    percentage of the valid samples in "field". Note, that you must detect the regime
    changepoints prior to calling this function. (They are expected to be stored
    parameter `cluster_field`.)

    Note, that it is possible to perform hypothesis tests for regime equality by
    passing the metric a function for p-value calculation and selecting linkage
    method "complete".

    Parameters
    ----------
    data : dios.DictOfSeries
        Data to process
    field : str
        Name of the column to process
    flags : saqc.Flags
        Container to store flags of the data.
    cluster_field : str
        Column in data, holding the cluster labels for the samples in field.
        (has to be indexed equal to field)
    spread : float
        A threshold denoting the value level, up to wich clusters a agglomerated.
    method : {"single", "complete", "average", "weighted", "centroid", "median", "ward"}, default "single"
        The linkage method for hierarchical (agglomerative) clustering of the variables.
    metric : Callable, default lambda x,y: np.abs(np.nanmean(x) - np.nanmean(y))
        A metric function for calculating the dissimilarity between 2 regimes.
        Defaults to the difference in mean.
    frac : float
        Has to be in [0,1]. Determines the minimum percentage of samples,
        the "normal" group has to comprise to be the normal group actually.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The flags object, holding flags and additional informations related to `data`.
        Flags values may have changed, relatively to the flags input.
    """
    reserverd = ["set_cluster", "set_flags", "flag"]
    kwargs = filterKwargs(kwargs, reserverd)
    return _assignRegimeAnomaly(
        data=data,
        field=field,
        flags=flags,
        cluster_field=cluster_field,
        spread=spread,
        method=method,
        metric=metric,
        frac=frac,
        **kwargs,
        # control args
        set_cluster=True,
        set_flags=False,
    )


def _assignRegimeAnomaly(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    cluster_field: str,
    spread: float,
    method: LinkageString = "single",
    metric: Callable[[np.array, np.array], float] = lambda x, y: np.abs(
        np.nanmean(x) - np.nanmean(y)
    ),
    frac: float = 0.5,
    set_cluster: bool = True,
    set_flags: bool = False,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    series = data[cluster_field]
    cluster = np.unique(series)
    cluster_dios = DictOfSeries({i: data[field][series == i] for i in cluster})
    plateaus = detectDeviants(cluster_dios, metric, spread, frac, method, "samples")

    if set_flags:
        for p in plateaus:
            flags[cluster_dios.iloc[:, p].index, field] = flag

    if set_cluster:
        for p in plateaus:
            if cluster[p] > 0:
                series[series == cluster[p]] = -cluster[p]

    data[cluster_field] = series
    return data, flags



#!/usr/bin/env python


__author__ = "Bert Palm"
__email__ = "bert.palm@ufz.de"
__copyright__ = "Copyright 2020, Helmholtz-Zentrum für Umweltforschung GmbH - UFZ"


from pandas.core.common import (
    is_bool_indexer,
    is_null_slice,
)
from pandas.core.dtypes.common import (
    is_nested_list_like,
)
from pandas.api.types import (
    is_list_like,
    is_hashable,
    is_integer,
    is_dict_like,
    is_scalar,
    # Unlike the example says, return lists False, not True
    # >>is_iterator([1, 2, 3])
    # >>False
    is_iterator,
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from saqc.core.translation.basescheme import (
    TranslationScheme,
    FloatScheme,
    SimpleScheme,
)
from saqc.core.translation.positionalscheme import PositionalScheme
from saqc.core.translation.dmpscheme import DmpScheme



from .test_setup import *
from pandas.core.dtypes.common import is_scalar


@pytest.mark.parametrize(("idxer", "exp"), [("a", s1), ("c", s3)])
def test__getitem_single(dios_aligned, idxer, exp):
    di = dios_aligned[idxer]
    assert isinstance(di, pd.Series)
    assert (di == exp).all()


@pytest.mark.parametrize(
    "idxer",
    [
        "x",
        "2",
        1000,
        None,
    ],
)
def test__getitem_single_fail(dios_aligned, idxer):
    with pytest.raises((KeyError, ValueError)):
        di = dios_aligned[idxer]


@pytest.mark.parametrize("idxer", BASIC_INDEXER)
def test__getitem_(dios_aligned, idxer):
    di = dios_aligned[idxer]

    assert isinstance(di, DictOfSeries)


@pytest.mark.parametrize("idxer", BASIC_INDEXER_FAIL)
def test__getitem_fail(dios_aligned, idxer):
    with pytest.raises((ValueError, KeyError)):
        dios_aligned[idxer]


@pytest.mark.parametrize(
    ("idxer", "exp"),
    [
        (slice(None), [s1 == s1, s2 == s2, s3 == s3, s4 == s4]),
        (dios_aligned__() > 5, [s1 > 5, s2 > 5, s3 > 5, s4 > 5]),
    ],
)
def test__setitem_single(dios_aligned, idxer, exp):
    di = dios_aligned
    di[idxer] = 99
    for i, c in enumerate(di):
        assert ((di[c] == 99) == exp[i]).all()


@pytest.mark.parametrize("idxer", BASIC_INDEXER_FAIL)
def test__setitem__fail(dios_aligned, idxer):
    with pytest.raises((ValueError, KeyError, IndexError)):
        dios_aligned[idxer] = 99



from .lib import *
from .dios import *

__all__ = [
    "DictOfSeries",
    "to_dios",
    "pprint_dios",
    "IntItype",
    "FloatItype",
    "NumItype",
    "DtItype",
    "ObjItype",
    "ItypeWarning",
    "ItypeCastWarning",
    "ItypeCastError",
    "is_itype",
    "is_itype_subtype",
    "is_itype_like",
    "get_itype",
    "cast_to_itype",
    "CastPolicy",
    "Opts",
    "OptsFields",
    "OptsFields",
    "dios_options",
    "example_DictOfSeries",
]



from .test_setup import *
import pytest


def _test(res, exp):

    if isinstance(exp, pd.DataFrame):
        eq, msg = dios_eq_df(res, exp, with_msg=True)
        assert eq, msg

    else:
        assert type(exp) == type(res)

        if isinstance(exp, pd.Series):
            eq, msg = diosSeries_eq_dfSeries(res, exp, with_msg=True)
            assert eq, msg

        # scalars
        else:
            assert res == exp


@pytest.mark.parametrize("idxer", BASIC_INDEXER)
def test_dflike__get__(df_aligned, dios_aligned, idxer):
    print(idxer)
    exp = df_aligned[idxer]
    res = dios_aligned[idxer]
    _test(res, exp)


@pytest.mark.parametrize("locR", R_LOC_INDEXER)
@pytest.mark.parametrize("locC", C_LOC_INDEXER)
def test_dflike__get_loc__(df_aligned, dios_aligned, locR, locC):
    print(locR)
    print(locC)
    exp = df_aligned.loc[locR, locC]
    res = dios_aligned.loc[locR, locC]
    _test(res, exp)


@pytest.mark.parametrize("ilocR", R_iLOC_INDEXER)
@pytest.mark.parametrize("ilocC", C_iLOC_INDEXER)
def test_dflike__get_iloc__(df_aligned, dios_aligned, ilocR, ilocC):
    print(ilocR)
    print(ilocC)
    exp = df_aligned.iloc[ilocR, ilocC]
    res = dios_aligned.iloc[ilocR, ilocC]
    _test(res, exp)


VALS = [
    99,
]


@pytest.mark.parametrize("idxer", BASIC_INDEXER)
@pytest.mark.parametrize("val", VALS)
def test_dflike__set__(df_aligned, dios_aligned, idxer, val):
    print(idxer)
    exp = df_aligned
    res = dios_aligned
    # NOTE: two test fail, pandas bul***it
    #   df[:2]    -> select 2 rows
    #   df[:2]=99 -> set 3 rows, WTF ???
    exp[idxer] = val
    res[idxer] = val
    _test(res, exp)


@pytest.mark.parametrize("locR", R_LOC_INDEXER)
@pytest.mark.parametrize("locC", C_LOC_INDEXER)
@pytest.mark.parametrize("val", VALS)
def test_dflike__set_loc__(df_aligned, dios_aligned, locR, locC, val):
    print(locR)
    print(locC)
    exp = df_aligned
    res = dios_aligned
    exp.loc[locR, locC] = val
    res.loc[locR, locC] = val
    _test(res, exp)


@pytest.mark.parametrize("ilocR", R_iLOC_INDEXER)
@pytest.mark.parametrize("ilocC", C_iLOC_INDEXER)
@pytest.mark.parametrize("val", VALS)
def test_dflike__set_iloc__(df_aligned, dios_aligned, ilocR, ilocC, val):
    print(ilocR)
    print(ilocC)
    exp = df_aligned
    res = dios_aligned
    exp.iloc[ilocR, ilocC] = val
    res.iloc[ilocR, ilocC] = val
    _test(res, exp)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd
import dios
from pathlib import Path

from saqc.core.reader import fromConfig, readFile
from saqc.core.register import FUNC_MAP, register, flagging

from tests.common import initData, writeIO


@pytest.fixture
def data() -> dios.DictOfSeries:
    return initData(3)


def test_packagedConfig():

    path = Path(__file__).parents[2] / "ressources/data"

    config_path = path / "config.csv"
    data_path = path / "data.csv"

    data = pd.read_csv(
        data_path,
        index_col=0,
        parse_dates=True,
    )
    saqc = fromConfig(config_path, data)
    saqc.result._validate()


def test_variableRegex(data):

    header = f"varname;test"
    tests = [
        ("'.*'", data.columns),
        ("'var(1|2)'", [c for c in data.columns if c[-1] in ("1", "2")]),
        ("'var[12]'", [c for c in data.columns if c[-1] in ("1", "2")]),
        ('".*3"', [c for c in data.columns if c[-1] == "3"]),
    ]

    for regex, expected in tests:
        fobj = writeIO(header + "\n" + f"{regex} ; flagDummy()")
        saqc = fromConfig(fobj, data=data)
        result = [field for field, _ in saqc._called]
        assert np.all(result == expected)

    tests = [
        ("var[12]", ["var[12]"]),  # not quoted -> not a regex
    ]
    for regex, expected in tests:
        fobj = writeIO(header + "\n" + f"{regex} ; flagDummy()")
        with pytest.warns(RuntimeWarning):
            saqc = fromConfig(fobj, data=data)
        result = [field for field, _ in saqc._called]
        assert np.all(result == expected)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_inlineComments(data):
    """
    adresses issue #3
    """
    config = f"""
    varname ; test
    pre2        ; flagDummy() # test
    """

    saqc = fromConfig(writeIO(config), data)
    _, func = saqc._called[0]
    assert func[0] == FUNC_MAP["flagDummy"]


def test_configReaderLineNumbers():
    config = f"""
    varname ; test
    #temp1      ; flagDummy()
    pre1        ; flagDummy()
    pre2        ; flagDummy()
    SM          ; flagDummy()
    #SM         ; flagDummy()
    # SM1       ; flagDummy()

    SM1         ; flagDummy()
    """
    planned = readFile(writeIO(config))
    expected = [4, 5, 6, 10]
    assert (planned.index == expected).all()


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_configFile(data):

    # check that the reader accepts different whitespace patterns

    config = f"""
    varname ; test

    #temp1      ; flagDummy()
    pre1; flagDummy()
    pre2        ;flagDummy()
    SM          ; flagDummy()
    #SM         ; flagDummy()
    # SM1       ; flagDummy()

    SM1;flagDummy()
    """
    fromConfig(writeIO(config), data)


def test_configChecks(data):

    var1, _, var3, *_ = data.columns

    @flagging()
    def flagFunc(data, field, flags, arg, opt_arg=None, **kwargs):
        flags[:, field] = np.nan
        return data, flags

    header = f"varname;test"
    tests = [
        (f"{var1};flagFunc(mn=0)", TypeError),  # bad argument name
        (f"{var1};flagFunc()", TypeError),  # not enough arguments
        (f"{var3};flagNothing()", NameError),  # unknown function
        (f"{var1}; min", TypeError),  # not a function call
    ]

    for test, expected in tests:
        fobj = writeIO(header + "\n" + test)
        with pytest.raises(expected):
            fromConfig(fobj, data=data)


def test_supportedArguments(data):

    # test if the following function arguments
    # are supported (i.e. parsing does not fail)

    # TODO: necessary?

    @flagging()
    def func(data, field, flags, kwarg, **kwargs):
        flags[:, field] = np.nan
        return data, flags

    var1 = data.columns[0]

    header = f"varname;test"
    tests = [
        f"{var1};func(kwarg=NAN)",
        f"{var1};func(kwarg='str')",
        f"{var1};func(kwarg=5)",
        f"{var1};func(kwarg=5.5)",
        f"{var1};func(kwarg=-5)",
        f"{var1};func(kwarg=True)",
        f"{var1};func(kwarg=sum([1, 2, 3]))",
    ]

    for test in tests:
        fobj = writeIO(header + "\n" + test)
        fromConfig(fobj, data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import numpy as np
import pandas as pd

from saqc.core.flags import (
    Flags,
    History,
    UNFLAGGED,
    GOOD,
    DOUBTFUL,
    BAD,
)
from saqc.core.translation.basescheme import (
    TranslationScheme,
    ForwardMap,
    BackwardMap,
)


class PositionalScheme(TranslationScheme):

    """
    Implements the translation from and to the flagging scheme implemented by CHS
    """

    DFILTER_DEFAULT = DOUBTFUL + 1

    _FORWARD: ForwardMap = {
        -6: UNFLAGGED,
        -5: UNFLAGGED,
        -2: UNFLAGGED,
        0: UNFLAGGED,
        1: DOUBTFUL,
        2: BAD,
    }
    _BACKWARD: BackwardMap = {
        np.nan: 0,
        UNFLAGGED: 0,
        GOOD: 0,
        DOUBTFUL: 1,
        BAD: 2,
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)

    def forward(self, flags: pd.DataFrame) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """

        data = {}
        for field, field_flags in flags.items():

            # explode the flags into sperate columns and drop the leading `9`
            df = pd.DataFrame(
                field_flags.astype(str).str.slice(start=1).apply(tuple).tolist(),
                index=field_flags.index,
            ).astype(int)

            # the exploded values form the History of `field`
            fflags = super()._translate(df, self._FORWARD)
            field_history = History(field_flags.index)
            for _, s in fflags.items():
                field_history.append(s)
            data[str(field)] = field_history

        return Flags(data)

    def backward(self, flags: Flags, **kwargs) -> pd.DataFrame:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        pd.DataFrame
        """
        out = {}
        for field in flags.columns:
            thist = flags.history[field].hist.replace(self._BACKWARD).astype(int)
            # concatenate the single flag values
            ncols = thist.shape[-1]
            init = 9 * 10 ** ncols
            bases = 10 ** np.arange(ncols - 1, -1, -1)

            tflags = init + (thist * bases).sum(axis=1)
            out[field] = tflags

        return pd.DataFrame(out).fillna(-9999).astype(int)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import dtw

from saqc.constants import BAD
from saqc.core.register import flagging
from saqc.lib.tools import customRoller


# todo should we mask `reference` even if the func fail if reference has NaNs
@flagging()
def flagPatternByWavelet(
    data,
    field,
    flags,
    reference,
    widths=(1, 2, 4, 8),
    waveform="mexh",
    flag=BAD,
    **kwargs,
):
    """
    Pattern recognition via wavelets.

    The steps are:
     1. work on chunks returned by a moving window
     2. each chunk is compared to the given pattern, using the wavelet algorithm as
        presented in [1]
     3. if the compared chunk is equal to the given pattern it gets flagged

    Parameters
    ----------

    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the data column, you want to correct.

    flags : saqc.Flags
        The flags belonging to ``data``.

    reference: str
        The fieldname in ``data`' which holds the pattern.

    widths: tuple of int
        Widths for wavelet decomposition. [1] recommends a dyadic scale.
        Default: (1,2,4,8)

    waveform: str
        Wavelet to be used for decomposition. Default: 'mexh'. See [2] for a list.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.

    flags : saqc.Flags
        The flags belongiong to `data`.

    References
    ----------

    The underlying pattern recognition algorithm using wavelets is documented here:
    [1] Maharaj, E.A. (2002): Pattern Recognition of Time Series using Wavelets. In: Härdle W., Rönz B. (eds) Compstat. Physica, Heidelberg, 978-3-7908-1517-7.

    The documentation of the python package used for the wavelt decomposition can be found here:
    [2] https://pywavelets.readthedocs.io/en/latest/ref/cwt.html#continuous-wavelet-families
    """

    dat = data[field]
    ref = data[reference].to_numpy()
    cwtmat_ref, _ = pywt.cwt(ref, widths, waveform)
    wavepower_ref = np.power(cwtmat_ref, 2)
    len_width = len(widths)
    sz = len(ref)

    assert len_width
    assert sz

    def func(x, y):
        return x.sum() / y.sum()

    def pvalue(chunk):
        cwtmat_chunk, _ = pywt.cwt(chunk, widths, waveform)
        wavepower_chunk = np.power(cwtmat_chunk, 2)

        # Permutation test on Powersum of matrix
        for i in range(len_width):
            x = wavepower_ref[i]
            y = wavepower_chunk[i]
            pval = permutation_test(
                x, y, method="approximate", num_rounds=200, func=func, seed=0
            )
            pval = min(pval, 1 - pval)
        return pval  # noqa # existence ensured by assert

    rolling = customRoller(dat, window=sz, min_periods=sz, forward=True)
    pvals = rolling.apply(pvalue, raw=True)
    markers = pvals > 0.01  # nans -> False

    # the markers are set on the left edge of the window. thus we must propagate
    # `sz`-many True's to the right of every marker.
    rolling = customRoller(markers, window=sz, min_periods=sz)
    mask = rolling.sum().fillna(0).astype(bool)

    flags[mask, field] = flag
    return data, flags


def calculateDistanceByDTW(
    data: pd.Series, reference: pd.Series, forward=True, normalize=True
):
    """
    Calculate the DTW-distance of data to pattern in a rolling calculation.

    The data is compared to pattern in a rolling window.
    The size of the rolling window is determined by the timespan defined
    by the first and last timestamp of the reference data's datetime index.

    For details see the linked functions in the `See Also` section.

    Parameters
    ----------
    data : pd.Series
        Data series. Must have datetime-like index, and must be regularly sampled.

    reference : : pd.Series
        Reference series. Must have datetime-like index, must not contain NaNs
        and must not be empty.

    forward: bool, default True
        If `True`, the distance value is set on the left edge of the data chunk. This
        means, with a perfect match, `0.0` marks the beginning of the pattern in
        the data. If `False`, `0.0` would mark the end of the pattern.

    normalize : bool, default True
        If `False`, return unmodified distances.
        If `True`, normalize distances by the number of observations in the reference.
        This helps to make it easier to find a good cutoff threshold for further
        processing. The distances then refer to the mean distance per datapoint,
        expressed in the datas units.

    Returns
    -------
    distance : pd.Series

    Notes
    -----
    The data must be regularly sampled, otherwise a ValueError is raised.
    NaNs in the data will be dropped before dtw distance calculation.

    See Also
    --------
    flagPatternByDTW : flag data by DTW
    """
    if reference.hasnans or reference.empty:
        raise ValueError("reference must not have nan's and must not be empty.")

    winsz = reference.index.max() - reference.index.min()
    reference = reference.to_numpy()

    def isPattern(chunk):
        return dtw.accelerated_dtw(chunk, reference, "euclidean")[0]

    # generate distances, excluding NaNs
    rolling = customRoller(
        data.dropna(), window=winsz, forward=forward, expand=False, closed="both"
    )
    distances: pd.Series = rolling.apply(isPattern, raw=True)

    if normalize:
        distances /= len(reference)

    return distances.reindex(index=data.index)  # reinsert NaNs


# todo should we mask `reference` even if the func fail if reference has NaNs
@flagging()
def flagPatternByDTW(
    data,
    field,
    flags,
    reference,
    max_distance=0.0,
    normalize=True,
    plot=False,
    flag=BAD,
    **kwargs,
):
    """Pattern Recognition via Dynamic Time Warping.

    The steps are:
     1. work on a moving window
     2. for each data chunk extracted from each window, a distance to the given pattern
        is calculated, by the dynamic time warping algorithm [1]
     3. if the distance is below the threshold, all the data in the window gets flagged

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The name of the data column

    flags : saqc.Flags
        The flags belonging to `data`.

    reference : str
        The name in `data` which holds the pattern. The pattern must not have NaNs,
        have a datetime index and must not be empty.

    max_distance : float, default 0.0
        Maximum dtw-distance between chunk and pattern, if the distance is lower than
        ``max_distance`` the data gets flagged. With default, ``0.0``, only exact
        matches are flagged.

    normalize : bool, default True
        If `False`, return unmodified distances.
        If `True`, normalize distances by the number of observations of the reference.
        This helps to make it easier to find a good cutoff threshold for further
        processing. The distances then refer to the mean distance per datapoint,
        expressed in the datas units.

    plot: bool, default False
        Show a calibration plot, which can be quite helpful to find the right threshold
        for `max_distance`. It works best with `normalize=True`. Do not use in automatic
        setups / pipelines. The plot show three lines:

        - data: the data the function was called on
        - distances: the calculated distances by the algorithm
        - indicator: have to distinct levels: `0` and the value of `max_distance`.
          If `max_distance` is `0.0` it defaults to `1`. Everywhere where the
          indicator is not `0` the data will be flagged.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.

    flags : saqc.Flags
        The flags belonging to `data`.

    Notes
    -----
    The window size of the moving window is set to equal the temporal extension of the
    reference datas datetime index.

    References
    ----------
    Find a nice description of underlying the Dynamic Time Warping Algorithm here:

    [1] https://cran.r-project.org/web/packages/dtw/dtw.pdf
    """
    ref = data[reference]
    dat = data[field]

    distances = calculateDistanceByDTW(dat, ref, forward=True, normalize=normalize)
    winsz = ref.index.max() - ref.index.min()

    # prevent nan propagation
    distances = distances.fillna(max_distance + 1)

    # find minima filter by threshold
    fw = customRoller(distances, window=winsz, forward=True, closed="both", expand=True)
    bw = customRoller(distances, window=winsz, closed="both", expand=True)
    minima = (fw.min() == bw.min()) & (distances <= max_distance)

    # Propagate True's to size of pattern.
    rolling = customRoller(minima, window=winsz, closed="both", expand=True)
    mask = rolling.sum() > 0

    if plot:
        df = pd.DataFrame()
        df["data"] = dat
        df["distances"] = distances
        df["indicator"] = mask.astype(float) * (max_distance or 1)
        df.plot()

    flags[mask, field] = flag
    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import warnings
from typing import Optional, Union, Tuple, Sequence, Callable
from typing_extensions import Literal

import numba
import numpy as np
import numpy.polynomial.polynomial as poly
import pandas as pd

from dios import DictOfSeries
from outliers import smirnov_grubbs
from scipy.optimize import curve_fit

from saqc.constants import *
from saqc.core import register, Flags
from saqc.core.register import flagging
from saqc.lib.tools import customRoller, findIndex, getFreqDelta, toSequence
from saqc.funcs.scores import assignKNNScore
from saqc.funcs.tools import copyField, dropField
from saqc.funcs.transformation import transform
import saqc.lib.ts_operators as ts_ops


@flagging()
def flagByStray(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: Optional[Union[int, str]] = None,
    min_periods: int = 11,
    iter_start: float = 0.5,
    alpha: float = 0.05,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flag outliers in 1-dimensional (score) data with the STRAY Algorithm.

    Find more information on the algorithm in References [1].

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store quality flags to data.

    freq : str, int, or None, default None
        Determines the segmentation of the data into partitions, the kNN algorithm is
        applied onto individually.

        * ``np.inf``: Apply Scoring on whole data set at once
        * ``x`` > 0 : Apply scoring on successive data chunks of periods length ``x``
        * Offset String : Apply scoring on successive partitions of temporal extension
          matching the passed offset string

    min_periods : int, default 11
        Minimum number of periods per partition that have to be present for a valid
        outlier dettection to be made in this partition. (Only of effect, if `freq`
        is an integer.) Partition min value must always be greater then the
        nn_neighbors value.

    iter_start : float, default 0.5
        Float in [0,1] that determines which percentage of data is considered
        "normal". 0.5 results in the stray algorithm to search only the upper 50 % of
        the scores for the cut off point. (See reference section for more information)

    alpha : float, default 0.05
        Level of significance by which it is tested, if a score might be drawn from
        another distribution, than the majority of the data.

    flag : float, default BAD
        flag to set.

    References
    ----------
    [1] Talagala, P. D., Hyndman, R. J., & Smith-Miles, K. (2019). Anomaly detection in
        high dimensional data. arXiv preprint arXiv:1908.04000.
    """
    scores = data[field].dropna()

    if scores.empty:
        return data, flags

    if not freq:
        freq = scores.shape[0]

    if isinstance(freq, str):
        partitions = scores.groupby(pd.Grouper(freq=freq))

    else:
        grouper_series = pd.Series(
            data=np.arange(0, scores.shape[0]), index=scores.index
        )
        grouper_series = grouper_series.transform(lambda x: int(np.floor(x / freq)))
        partitions = scores.groupby(grouper_series)

    # calculate flags for every partition
    for _, partition in partitions:

        if partition.empty | (partition.shape[0] < min_periods):
            continue

        sample_size = partition.shape[0]

        sorted_i = partition.values.argsort()
        resids = partition.values[sorted_i]
        gaps = np.append(0, np.diff(resids))

        tail_size = int(max(min(50, np.floor(sample_size / 4)), 2))
        tail_indices = np.arange(2, tail_size + 1)

        i_start = int(max(np.floor(sample_size * iter_start), 1) + 1)
        ghat = np.array([np.nan] * sample_size)

        for i in range(i_start - 1, sample_size):
            ghat[i] = sum((tail_indices / (tail_size - 1)) * gaps[i - tail_indices + 1])

        log_alpha = np.log(1 / alpha)
        for iter_index in range(i_start - 1, sample_size):
            if gaps[iter_index] > log_alpha * ghat[iter_index]:
                index = partition.index[sorted_i[iter_index:]]
                flags[index, field] = flag
                break

    return data, flags


def _evalStrayLabels(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    fields: Sequence[str],
    reduction_range: Optional[str] = None,
    reduction_drop_flagged: bool = False,  # TODO: still a case ?
    reduction_thresh: float = 3.5,
    reduction_min_periods: int = 1,
    at_least_one: bool = True,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function "reduces" an observations flag to components of it, by applying MAD
    (See references) test onto every components temporal surrounding.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the column, holding the labels to be evaluated.

    flags : saqc.Flags
        Container to store quality flags to data.

    fields : list of str
        A list of strings, holding the column names of the variables, the stray labels
        shall be projected onto.

    val_frame : (N,M) pd.DataFrame
        Input NxM DataFrame of observations, where N is the number of observations and
        M the number of components per observation.

    to_flag_frame : pandas.DataFrame
        Input dataframe of observations to be tested, where N is the number of
        observations and M the number of components per observation.

    reduction_range : {None, str}
        An offset string, denoting the range of the temporal surrounding to include
        into the MAD testing. If ``None`` is passed, no testing will be performed and
        all fields will have the stray flag projected.

    reduction_drop_flagged : bool, default False
        Wheather or not to drop flagged values other than the value under test, from the
        temporal surrounding before checking the value with MAD.

    reduction_thresh : float, default 3.5
        The `critical` value, controlling wheather the MAD score is considered
        referring to an outlier or not. Higher values result in less rigid flagging.
        The default value is widely used in the literature. See references section
        for more details ([1]).

    at_least_one : bool, default True
        If none of the variables, the outlier label shall be reduced to, is an outlier
        with regard to the test, all (True) or none (False) of the variables are flagged

    flag : float, default BAD
        flag to set.

    References
    ----------
    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    """
    val_frame = data[fields].to_df()
    stray_detects = flags[field] > UNFLAGGED
    stray_detects = stray_detects[stray_detects]
    to_flag_frame = pd.DataFrame(False, columns=fields, index=stray_detects.index)

    if reduction_range is None:
        for field in to_flag_frame.columns:
            flags[to_flag_frame.index, field] = flag
        return data, flags

    for var in fields:
        for index in enumerate(to_flag_frame.index):

            index_slice = slice(
                index[1] - pd.Timedelta(reduction_range),
                index[1] + pd.Timedelta(reduction_range),
            )
            test_slice = val_frame[var][index_slice].dropna()

            # check, wheather value under test is sufficiently centered:
            first = test_slice.first_valid_index()
            last = test_slice.last_valid_index()
            min_range = pd.Timedelta(reduction_range) / 4

            if (
                pd.Timedelta(index[1] - first) < min_range
                or pd.Timedelta(last - index[1]) < min_range
            ):
                polydeg = 0
            else:
                polydeg = 2

            if reduction_drop_flagged:
                test_slice = test_slice.drop(to_flag_frame.index, errors="ignore")

            if test_slice.shape[0] < reduction_min_periods:
                to_flag_frame.loc[index[1], var] = True
                continue

            x = test_slice.index.values.astype(float)
            x_0 = x[0]
            x = (x - x_0) / 10 ** 12

            polyfitted = poly.polyfit(y=test_slice.values, x=x, deg=polydeg)

            testval = poly.polyval(
                (float(index[1].to_numpy()) - x_0) / 10 ** 12, polyfitted
            )
            testval = val_frame[var][index[1]] - testval

            resids = test_slice.values - poly.polyval(x, polyfitted)
            med_resids = np.median(resids)
            MAD = np.median(np.abs(resids - med_resids))
            crit_val = 0.6745 * (abs(med_resids - testval)) / MAD

            if crit_val > reduction_thresh:
                to_flag_frame.loc[index[1], var] = True

    if at_least_one:
        to_flag_frame[~to_flag_frame.any(axis=1)] = True

    for field in to_flag_frame.columns:
        col = to_flag_frame[field]
        flags[col[col].index, field] = flag

    return data, flags


def _expFit(
    val_frame,
    scoring_method="kNNMaxGap",
    n_neighbors=10,
    iter_start=0.5,
    alpha=0.05,
    bin_frac=10,
):
    """
    Find outliers in multi dimensional observations.

    The general idea is to assigning scores to every observation based on the observations neighborhood in the space
    of observations. Then, the gaps between the (greatest) scores are tested for beeing drawn from the same
    distribution, as the majority of the scores.

    Note, that no normalizations/transformations are applied to the different components (data columns)
    - those are expected to be applied previously, if necessary.

    Parameters
    ----------
    val_frame : (N,M) ndarray
        Input NxM array of observations, where N is the number of observations and M the number of components per
        observation.
    scoring_method : {'kNNSum', 'kNNMaxGap'}, default 'kNNMaxGap'
        Scoring method applied.
        `'kNNSum'`: Assign to every point the sum of the distances to its 'n' nearest neighbors.
        `'kNNMaxGap'`: Assign to every point the distance to the neighbor with the "maximum gap" to its predecessor
        in the hierarchy of the `n` nearest neighbors. (see reference section for further descriptions)
    n_neighbors : int, default 10
        Number of neighbors included in the scoring process for every datapoint.
    iter_start : float, default 0.5
        Float in [0,1] that determines which percentage of data is considered "normal". 0.5 results in the expfit
        algorithm to search only the upper 50 % of the scores for the cut off point. (See reference section for more
        information)
    alpha : float, default 0.05
        Niveau of significance by which it is tested, if a score might be drawn from another distribution, than the
        majority of the data.
    bin_frac : {int, str}, default 10
        Controls the binning for the histogram in the fitting step. If an integer is passed, the residues will
        equidistantly be covered by `bin_frac` bins, ranging from the minimum to the maximum of the residues.
        If a string is passed, it will be passed on to the ``numpy.histogram_bin_edges`` method.
    """

    kNNfunc = getattr(ts_ops, scoring_method)
    resids = kNNfunc(val_frame.values, n_neighbors=n_neighbors, algorithm="ball_tree")
    data_len = resids.shape[0]

    # sorting
    sorted_i = resids.argsort()
    resids = resids[sorted_i]
    iter_index = int(np.floor(resids.size * iter_start))
    # initialize condition variables:
    crit_val = np.inf
    test_val = 0
    neg_log_alpha = -np.log(alpha)

    # define exponential dist density function:
    def fit_function(x, lambd):
        return lambd * np.exp(-lambd * x)

    # initialise sampling bins
    if isinstance(bin_frac, int):
        binz = np.linspace(
            resids[0], resids[-1], 10 * int(np.ceil(data_len / bin_frac))
        )
    elif bin_frac in [
        "auto",
        "fd",
        "doane",
        "scott",
        "stone",
        "rice",
        "sturges",
        "sqrt",
    ]:
        binz = np.histogram_bin_edges(resids, bins=bin_frac)
    else:
        raise ValueError(f"Can't interpret {bin_frac} as an binning technique.")

    binzenters = np.array([0.5 * (binz[i] + binz[i + 1]) for i in range(len(binz) - 1)])
    # inititialize full histogram:
    full_hist, binz = np.histogram(resids, bins=binz)
    # check if start index is sufficiently high (pointing at resids value beyond histogram maximum at least):
    hist_argmax = full_hist.argmax()

    if hist_argmax >= findIndex(binz, resids[iter_index - 1], 0):
        raise ValueError(
            "Either the data histogram is too strangely shaped for oddWater OD detection - "
            "or a too low value for 'offset' was passed "
            "(offset better be much greater 0.5)"
        )
    # GO!
    iter_max_bin_index = findIndex(binz, resids[iter_index - 1], 0)
    upper_tail_index = int(np.floor(0.5 * hist_argmax + 0.5 * iter_max_bin_index))
    resids_tail_index = findIndex(resids, binz[upper_tail_index], 0)
    upper_tail_hist, bins = np.histogram(
        resids[resids_tail_index:iter_index],
        bins=binz[upper_tail_index : iter_max_bin_index + 1],
    )

    while (test_val < crit_val) & (iter_index < resids.size - 1):
        iter_index += 1
        new_iter_max_bin_index = findIndex(binz, resids[iter_index - 1], 0)
        # following if/else block "manually" expands the data histogram and circumvents calculation of the complete
        # histogram in any new iteration.
        if new_iter_max_bin_index == iter_max_bin_index:
            upper_tail_hist[-1] += 1
        else:
            upper_tail_hist = np.append(
                upper_tail_hist, np.zeros([new_iter_max_bin_index - iter_max_bin_index])
            )
            upper_tail_hist[-1] += 1
            iter_max_bin_index = new_iter_max_bin_index
            upper_tail_index_new = int(
                np.floor(0.5 * hist_argmax + 0.5 * iter_max_bin_index)
            )
            upper_tail_hist = upper_tail_hist[upper_tail_index_new - upper_tail_index :]
            upper_tail_index = upper_tail_index_new

        # fitting

        lambdA, _ = curve_fit(
            fit_function,
            xdata=binzenters[upper_tail_index:iter_max_bin_index],
            ydata=upper_tail_hist,
            p0=[-np.log(alpha / resids[iter_index])],
        )

        crit_val = neg_log_alpha / lambdA
        test_val = resids[iter_index]

    return val_frame.index[sorted_i[iter_index:]]


@register(
    mask=["field"],
    demask=["field"],
    squeeze=["field"],
    multivariate=True,
    handles_target=True,
)
def flagMVScores(
    data: DictOfSeries,
    field: Sequence[str],
    flags: Flags,
    trafo: Callable[[pd.Series], pd.Series] = lambda x: x,
    alpha: float = 0.05,
    n: int = 10,
    func: Callable[[pd.Series], float] = np.sum,
    iter_start: float = 0.5,
    partition: Optional[Union[int, str]] = None,
    partition_min: int = 11,
    stray_range: Optional[str] = None,
    drop_flagged: bool = False,  # TODO: still a case ?
    thresh: float = 3.5,
    min_periods: int = 1,
    target: str = None,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The algorithm implements a 3-step outlier detection procedure for simultaneously
    flagging of higher dimensional data (dimensions > 3).

    In references [1], the procedure is introduced and exemplified with an
    application on hydrological data. See the notes section for an overview over the
    algorithms basic steps.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : list of str
        List of fieldnames, corresponding to the variables that are to be included
        into the flagging process.

    flags : saqc.Flags
        Container to store quality flags to data.

    trafo : callable, default lambda x:x
        Transformation to be applied onto every column before scoring. Will likely
        get deprecated soon. Its better to transform the data in a processing step,
        preceeeding the call to ``flagMVScores``.

    alpha : float, default 0.05
        Level of significance by which it is tested, if an observations score might
        be drawn from another distribution than the majority of the observation.

    n : int, default 10
        Number of neighbors included in the scoring process for every datapoint.

    func : Callable[numpy.array, float], default np.sum
        The function that maps the set of every points k-nearest neighbor distances
        onto a certain scoring.

    iter_start : float, default 0.5
        Float in [0,1] that determines which percentage of data is considered
        "normal". 0.5 results in the threshing algorithm to search only the upper 50
        % of the scores for the cut off point. (See reference section for more
        information)

    partition : {None, str, int}, default None
        Only effective when `threshing` = 'stray'. Determines the size of the data
        partitions, the data is decomposed into. Each partition is checked seperately
        for outliers. If a String is passed, it has to be an offset string and it
        results in partitioning the data into parts of according temporal length. If
        an integer is passed, the data is simply split up into continous chunks of
        `freq` periods. if ``None`` is passed (default), all the data will be tested
        in one run.

    partition_min : int, default 11
        Only effective when `threshing` = 'stray'. Minimum number of periods per
        partition that have to be present for a valid outlier detection to be made in
        this partition. (Only of effect, if `stray_partition` is an integer.)

    partition_trafo : bool, default True
        Whether or not to apply the passed transformation on every partition the
        algorithm is applied on, separately.

    stray_range : {None, str}, default None
        If not None, it is tried to reduce the stray result onto single outlier
        components of the input fields. An offset string, denoting the range of the
        temporal surrounding to include into the MAD testing while trying to reduce
        flags.

    drop_flagged : bool, default False
        Only effective when `range` is not ``None``. Whether or not to drop flagged
        values other than the value under test from the temporal surrounding before
        checking the value with MAD.

    thresh : float, default 3.5
        Only effective when `range` is not ``None``. The `critical` value,
        controlling wheather the MAD score is considered referring to an outlier or
        not. Higher values result in less rigid flagging. The default value is widely
        considered apropriate in the literature.

    min_periods : int, 1
        Only effective when `range` is not ``None``. Minimum number of meassurements
        necessarily present in a reduction interval for reduction actually to be
        performed.

    target : None
        ignored.

    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed, relatively to the flags input.

    Notes
    -----
    The basic steps are:

    1. transforming

    The different data columns are transformed via timeseries transformations to
    (a) make them comparable and
    (b) make outliers more stand out.

    This step is usually subject to a phase of research/try and error. See [1] for more
    details.

    Note, that the data transformation as an built-in step of the algorithm,
    will likely get deprecated soon. Its better to transform the data in a processing
    step, preceeding the multivariate flagging process. Also, by doing so, one gets
    mutch more control and variety in the transformation applied, since the `trafo`
    parameter only allows for application of the same transformation to all of the
    variables involved.

    2. scoring

    Every observation gets assigned a score depending on its k nearest neighbors. See
    the `scoring_method` parameter description for details on the different scoring
    methods. Furthermore [1], [2] may give some insight in the pro and cons of the
    different methods.

    3. threshing

    The gaps between the (greatest) scores are tested for beeing drawn from the same
    distribution as the majority of the scores. If a gap is encountered, that,
    with sufficient significance, can be said to not be drawn from the same
    distribution as the one all the smaller gaps are drawn from, than the observation
    belonging to this gap, and all the observations belonging to gaps larger then
    this gap, get flagged outliers. See description of the `threshing` parameter for
    more details. Although [2] gives a fully detailed overview over the `stray`
    algorithm.
    """
    if target:
        raise NotImplementedError("target is not implemented for this function")

    fields = toSequence(field)
    for f in fields:
        data, flags = copyField(data, f, flags, f"trafo_{f}")
        data, flags = transform(data, f"trafo_{f}", flags, func=trafo, freq=partition)

    data, flags = assignKNNScore(
        data=data,
        field=[f"trafo_{f}" for f in fields],
        flags=flags,
        target="kNN",
        n=n,
        func=func,
        freq=partition,
        method="ball_tree",
        min_periods=partition_min,
        **kwargs,
    )

    data, flags = flagByStray(
        data,
        "kNN",
        flags,
        freq=partition,
        min_periods=partition_min,
        iter_start=iter_start,
        alpha=alpha,
        flag=flag,
        **kwargs,
    )

    data, flags = _evalStrayLabels(
        data=data,
        field="kNN",
        fields=fields,
        flags=flags,
        reduction_range=stray_range,
        reduction_drop_flagged=drop_flagged,
        reduction_thresh=thresh,
        reduction_min_periods=min_periods,
        flag=flag,
        **kwargs,
    )
    data, flags = dropField(data, "kNN", flags)
    for f in fields:
        data, flags = dropField(data, f"trafo_{f}", flags)

    return data, flags


@flagging()
def flagRaise(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    thresh: float,
    raise_window: str,
    freq: str,
    average_window: Optional[str] = None,
    raise_factor: float = 2.0,
    slope: Optional[float] = None,
    weight: float = 0.8,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags raises and drops in value courses, that exceed a certain threshold
    within a certain timespan.

    The parameter variety of the function is owned to the intriguing
    case of values, that "return" from outlierish or anomalious value levels and
    thus exceed the threshold, while actually being usual values.

    NOTE, the dataset is NOT supposed to be harmonized to a time series with an
    equidistant frequency grid.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store flags of the data.
    thresh : float
        The threshold, for the total rise (thresh > 0), or total drop (thresh < 0),
        value courses must not exceed within a timespan of length `raise_window`.
    raise_window : str
        An offset string, determining the timespan, the rise/drop thresholding refers
        to. Window is inclusively defined.
    freq : str
        An offset string, determining The frequency, the timeseries to-be-flagged is
        supposed to be sampled at. The window is inclusively defined.
    average_window : {None, str}, default None
        See condition (2) of the description linked in the references. Window is
        inclusively defined. The window defaults to 1.5 times the size of `raise_window`
    raise_factor : float, default 2
        See second condition listed in the notes below.
    slope : {None, float}, default None
        See third condition listed in the notes below.
    weight : float, default 0.8
        See third condition listed in the notes below.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed, relatively to the flags input.

    Notes
    -----
    The value :math:`x_{k}` of a time series :math:`x` with associated
    timestamps :math:`t_i`, is flagged a raise, if:

    * There is any value :math:`x_{s}`, preceeding :math:`x_{k}` within `raise_window`
      range, so that:

      * :math:`M = |x_k - x_s | >`  `thresh` :math:`> 0`

    * The weighted average :math:`\\mu^{*}` of the values, preceding :math:`x_{k}`
      within `average_window`
      range indicates, that :math:`x_{k}` does not return from an "outlierish" value
      course, meaning that:

      * :math:`x_k > \\mu^* + ( M` / `mean_raise_factor` :math:`)`

    * Additionally, if ``min_slope`` is not `None`, :math:`x_{k}` is checked for being
      sufficiently divergent from its very predecessor :math:`x_{k-1}`, meaning that, it
      is additionally checked if:

      * :math:`x_k - x_{k-1} >` `min_slope`
      * :math:`t_k - t_{k-1} >` `weight` :math:`\\times` `freq`

    """

    # prepare input args
    dataseries = data[field].dropna()
    raise_window = pd.Timedelta(raise_window)
    freq = pd.Timedelta(freq)
    if slope is not None:
        slope = np.abs(slope)

    if average_window is None:
        average_window = 1.5 * pd.Timedelta(raise_window)

    if thresh < 0:
        dataseries *= -1
        thresh *= -1

    def raise_check(x, thresh):
        test_set = x[-1] - x[0:-1]
        max_val = np.max(test_set)
        if max_val >= thresh:
            return max_val
        else:
            return np.nan

    def custom_rolling_mean(x):
        return np.sum(x[:-1])

    # get invalid-raise/drop mask:
    raise_series = dataseries.rolling(raise_window, min_periods=2, closed="both")

    numba_boost = True
    if numba_boost:
        raise_check = numba.jit(raise_check, nopython=True)
        raise_series = raise_series.apply(
            raise_check, args=(thresh,), raw=True, engine="numba"
        )
    else:
        raise_series = raise_series.apply(raise_check, args=(thresh,), raw=True)

    if raise_series.isna().all():
        return data, flags

    # "unflag" values of insufficient deviation to their predecessors
    if slope is not None:
        w_mask = (
            pd.Series(dataseries.index).diff().dt.total_seconds() / freq.total_seconds()
        ) > weight
        slope_mask = np.abs(dataseries.diff()) < slope
        to_unflag = raise_series.notna() & w_mask.values & slope_mask
        raise_series[to_unflag] = np.nan

    # calculate and apply the weighted mean weights (pseudo-harmonization):
    weights = (
        pd.Series(dataseries.index).diff(periods=2).shift(-1).dt.total_seconds()
        / freq.total_seconds()
        / 2
    )

    weights.iloc[0] = 0.5 + (
        dataseries.index[1] - dataseries.index[0]
    ).total_seconds() / (freq.total_seconds() * 2)

    weights.iloc[-1] = 0.5 + (
        dataseries.index[-1] - dataseries.index[-2]
    ).total_seconds() / (freq.total_seconds() * 2)

    weights[weights > 1.5] = 1.5
    weights.index = dataseries.index
    weighted_data = dataseries.mul(weights)

    # rolling weighted mean calculation
    weighted_rolling_mean = weighted_data.rolling(
        average_window, min_periods=2, closed="both"
    )
    weights_rolling_sum = weights.rolling(average_window, min_periods=2, closed="both")
    if numba_boost:
        custom_rolling_mean = numba.jit(custom_rolling_mean, nopython=True)
        weighted_rolling_mean = weighted_rolling_mean.apply(
            custom_rolling_mean, raw=True, engine="numba"
        )
        weights_rolling_sum = weights_rolling_sum.apply(
            custom_rolling_mean, raw=True, engine="numba"
        )
    else:
        weighted_rolling_mean = weighted_rolling_mean.apply(
            custom_rolling_mean, raw=True
        )
        weights_rolling_sum = weights_rolling_sum.apply(
            custom_rolling_mean, raw=True, engine="numba"
        )

    weighted_rolling_mean = weighted_rolling_mean / weights_rolling_sum
    # check means against critical raise value:
    to_flag = dataseries >= weighted_rolling_mean + (raise_series / raise_factor)
    to_flag &= raise_series.notna()
    flags[to_flag[to_flag].index, field] = flag

    return data, flags


@flagging()
def flagMAD(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: str,
    z: float = 3.5,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function represents an implementation of the modyfied Z-score outlier detection method.

    See references [1] for more details on the algorithm.

    Note, that the test needs the input data to be sampled regularly (fixed sampling rate).

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged. (Here a dummy, for structural reasons)
    flags : saqc.Flags
        Container to store flags of the data.
    window : str
       Offset string. Denoting the windows size that the "Z-scored" values have to lie in.
    z: float, default 3.5
        The value the Z-score is tested against. Defaulting to 3.5 (Recommendation of [1])
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed, relatively to the flags input.

    References
    ----------
    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    """
    d = data[field]
    if d.empty:
        return data, flags

    median = d.rolling(window=window, closed="both").median()
    diff = (d - median).abs()
    mad = diff.rolling(window=window, closed="both").median()
    mask = (mad > 0) & (0.6745 * diff > z * mad)
    # NOTE:
    # In pandas <= 0.25.3, the window size is not fixed if the
    # window-argument to rolling is a frequency. That implies,
    # that during the first iterations the window has a size of
    # 1, 2, 3, ... until it eventually covers the desired time
    # span. For stuff like the calculation of median, that is rather
    # unfortunate, as the size of the calculation base might differ
    # heavily. So don't flag something until, the window reaches
    # its target size
    if not isinstance(window, int):
        index = mask.index
        mask.loc[index < index[0] + pd.to_timedelta(window)] = False

    flags[mask, field] = flag
    return data, flags


@flagging()
def flagOffset(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    thresh: float,
    tolerance: float,
    window: Union[int, str],
    thresh_relative: Optional[float] = None,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    A basic outlier test that work on regular and irregular sampled data

    The test classifies values/value courses as outliers by detecting not only a rise
    in value, but also, checking for a return to the initial value level.

    Values :math:`x_n, x_{n+1}, .... , x_{n+k}` of a timeseries :math:`x` with
    associated timestamps :math:`t_n, t_{n+1}, .... , t_{n+k}` are considered spikes, if

    1. :math:`|x_{n-1} - x_{n + s}| >` `thresh`, for all :math:`s \\in [0,1,2,...,k]`

    2. :math:`|x_{n-1} - x_{n+k+1}| <` `tolerance`

    3. :math:`|t_{n-1} - t_{n+k+1}| <` `window`

    Note, that this definition of a "spike" not only includes one-value outliers, but
    also plateau-ish value courses.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The field in data.
    flags : saqc.Flags
        Container to store flags of the data.
    thresh : float
        Minimum difference between to values, to consider the latter one as a spike. See condition (1)
    tolerance : float
        Maximum difference between pre-spike and post-spike values. See condition (2)
    window : {str, int}, default '15min'
        Maximum length of "spiky" value courses. See condition (3). Integer defined window length are only allowed for
        regularly sampled timeseries.
    thresh_relative : {float, None}, default None
        Relative threshold.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed, relatively to the flags input.

    References
    ----------
    The implementation is a time-window based version of an outlier test from the UFZ Python library,
    that can be found here:

    https://git.ufz.de/chs/python/blob/master/ufz/level1/spike.py

    """
    dataseries = data[field].dropna()
    if dataseries.empty:
        return data, flags

    # using reverted series - because ... long story.
    ind = dataseries.index
    rev_ind = ind[0] + ((ind[-1] - ind)[::-1])
    map_i = pd.Series(ind, index=rev_ind)
    dataseries = pd.Series(dataseries.values, index=rev_ind)

    if isinstance(window, int):
        delta = getFreqDelta(dataseries.index)
        window = delta * window
        if not delta:
            raise TypeError(
                "Only offset string defined window sizes allowed for irrgegularily sampled timeseries"
            )

    # get all the entries preceding a significant jump
    if thresh:
        post_jumps = dataseries.diff().abs() > thresh

    if thresh_relative:
        s = np.sign(thresh_relative)
        rel_jumps = s * (dataseries.shift(1) - dataseries).div(dataseries.abs()) > abs(
            thresh_relative
        )
        if thresh:
            post_jumps = rel_jumps & post_jumps
        else:
            post_jumps = rel_jumps

    post_jumps = post_jumps[post_jumps]
    if post_jumps.empty:
        return data, flags

    # get all the entries preceding a significant jump
    # and its successors within "length" range
    to_roll = post_jumps.reindex(
        dataseries.index, method="bfill", tolerance=window, fill_value=False
    ).dropna()
    to_roll = dataseries[to_roll]

    if thresh_relative:

        def spikeTester(chunk, thresh=abs(thresh_relative), tol=tolerance):
            jump = chunk[-2] - chunk[-1]
            thresh = thresh * abs(jump)
            chunk_stair = (np.sign(jump) * (chunk - chunk[-1]) < thresh)[::-1].cumsum()
            initial = np.searchsorted(chunk_stair, 2)
            if initial == len(chunk):
                return 0
            if np.abs(chunk[-initial - 1] - chunk[-1]) < tol:
                return initial - 1
            return 0

    else:

        # define spike testing function to roll with (no  rel_check):
        def spikeTester(chunk, thresh=thresh, tol=tolerance):
            # signum change!!!
            chunk_stair = (
                np.sign(chunk[-2] - chunk[-1]) * (chunk - chunk[-1]) < thresh
            )[::-1].cumsum()
            initial = np.searchsorted(chunk_stair, 2)
            if initial == len(chunk):
                return 0
            if np.abs(chunk[-initial - 1] - chunk[-1]) < tol:
                return initial - 1
            return 0

    roller = customRoller(to_roll, window=window, min_periods=2, closed="both")
    engine = None if len(to_roll) < 200000 else "numba"
    result = roller.apply(spikeTester, raw=True, engine=engine)

    ignore = pd.Series(True, index=to_roll.index)
    ignore[post_jumps.index] = False
    result[ignore] = np.nan

    result.index = map_i[result.index]

    # correct the result: only those values define plateaus, that do not have
    # values at their left starting point, that belong to other plateaus themself:
    def calcResult(result):
        var_num = result.shape[0]
        flag_scopes = np.zeros(var_num, dtype=bool)
        for k in range(var_num):
            if result[k] > 0:
                k_r = int(result[k])
                # validity check: plateuas start isnt another plateaus end:
                if not flag_scopes[k - k_r - 1]:
                    flag_scopes[(k - k_r) : k] = True
        return pd.Series(flag_scopes, index=result.index)

    cresult = calcResult(result)
    cresult = cresult[cresult].index
    flags[cresult, field] = flag
    return data, flags


@flagging()
def flagByGrubbs(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    alpha: float = 0.05,
    min_periods: int = 8,
    pedantic: bool = False,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags values that are regarded outliers due to the grubbs test.

    See reference [1] for more information on the grubbs tests definition.

    The (two-sided) test gets applied onto data chunks of size "window". The tests
    application  will be iterated on each data-chunk under test, till no more
    outliers are detected in that chunk.

    Note, that the test performs poorely for small data chunks (resulting in heavy
    overflagging). Therefor you should select "window" so that every window contains
    at least > 8 values and also adjust the min_periods values accordingly.

    Note, that the data to be tested by the grubbs test are expected to be distributed
    "normalish".

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store flags of the data.
    window : {int, str}
        The size of the window you want to use for outlier testing. If an integer is
        passed, the size refers to the number of periods of every testing window. If a
        string is passed, it has to be an offset string, and will denote the total
        temporal extension of every window.
    alpha : float, default 0.05
        The level of significance, the grubbs test is to be performed at. (between 0 and 1)
    min_periods : int, default 8
        The minimum number of values that have to be present in an interval under test,
        for a grubbs test result to be accepted. Only makes sence in case `window` is
        an offset string.
    pedantic: boolean, default False
        If True, every value gets checked twice for being an outlier. Ones in the
        initial rolling window and one more time in a rolling window that is lagged
        by half the windows delimeter (window/2). Recommended for avoiding false
        positives at the window edges. Only available when rolling with integer
        defined window size.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the flags input.

    References
    ----------
    introduction to the grubbs test:

    [1] https://en.wikipedia.org/wiki/Grubbs%27s_test_for_outliers
    """
    datcol = data[field].copy()
    rate = getFreqDelta(datcol.index)

    # if timeseries that is analyzed, is regular,
    # window size can be transformed to a number of periods:
    if rate and isinstance(window, str):
        window = pd.Timedelta(window) // rate

    to_group = pd.DataFrame(data={"ts": datcol.index, "data": datcol})
    to_flag = pd.Series(False, index=datcol.index)

    # period number defined test intervals
    if isinstance(window, int):
        grouper_series = pd.Series(
            data=np.arange(0, datcol.shape[0]), index=datcol.index
        )
        grouper_series_lagged = grouper_series + (window / 2)
        grouper_series = grouper_series.transform(lambda x: x // window)
        grouper_series_lagged = grouper_series_lagged.transform(lambda x: x // window)
        partitions = to_group.groupby(grouper_series)
        partitions_lagged = to_group.groupby(grouper_series_lagged)

    # offset defined test intervals:
    else:
        partitions = to_group.groupby(pd.Grouper(freq=window))
        partitions_lagged = []

    for _, partition in partitions:
        if partition.shape[0] > min_periods:
            detected = smirnov_grubbs.two_sided_test_indices(
                partition["data"].values, alpha=alpha
            )
            detected = partition["ts"].iloc[detected]
            to_flag[detected.index] = True

    if isinstance(window, int) and pedantic:
        to_flag_lagged = pd.Series(False, index=datcol.index)

        for _, partition in partitions_lagged:
            if partition.shape[0] > min_periods:
                detected = smirnov_grubbs.two_sided_test_indices(
                    partition["data"].values, alpha=alpha
                )
                detected = partition["ts"].iloc[detected]
                to_flag_lagged[detected.index] = True

        to_flag &= to_flag_lagged

    flags[to_flag, field] = flag
    return data, flags


@flagging()
def flagRange(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    min: float = -np.inf,
    max: float = np.inf,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function flags values not covered by the closed interval [`min`, `max`].

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store flags of the data.
    min : float
        Lower bound for valid data.
    max : float
        Upper bound for valid data.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
    """

    # using .values is much faster
    datacol = data[field].values
    mask = (datacol < min) | (datacol > max)
    flags[mask, field] = flag
    return data, flags


@register(
    mask=["field"],
    demask=["field"],
    squeeze=["field"],
    multivariate=True,
    handles_target=True,
)
def flagCrossStatistic(
    data: DictOfSeries,
    field: Sequence[str],
    flags: Flags,
    thresh: float,
    method: Literal["modZscore", "Zscore"] = "modZscore",
    flag: float = BAD,
    target: str = None,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function checks for outliers relatively to the "horizontal" input data axis.

    For `fields` :math:`=[f_1,f_2,...,f_N]` and timestamps :math:`[t_1,t_2,...,t_K]`, the following steps are taken
    for outlier detection:

    1. All timestamps :math:`t_i`, where there is one :math:`f_k`, with :math:`data[f_K]` having no entry at
       :math:`t_i`, are excluded from the following process (inner join of the :math:`f_i` fields.)
    2. for every :math:`0 <= i <= K`, the value
       :math:`m_j = median(\\{data[f_1][t_i], data[f_2][t_i], ..., data[f_N][t_i]\\})` is calculated
    3. for every :math:`0 <= i <= K`, the set
       :math:`\\{data[f_1][t_i] - m_j, data[f_2][t_i] - m_j, ..., data[f_N][t_i] - m_j\\}` is tested for outliers with the
       specified method (`cross_stat` parameter).

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : list of str
        List of fieldnames in data, determining wich variables are to be included into the flagging process.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    thresh : float
        Threshold which the outlier score of an value must exceed, for being flagged an outlier.
    method : {'modZscore', 'Zscore'}, default 'modZscore'
        Method used for calculating the outlier scores.

        * ``'modZscore'``: Median based "sigma"-ish approach. See Referenecs [1].
        * ``'Zscore'``: Score values by how many times the standard deviation they differ from the median.
          See References [1]

    flag : float, default BAD
        flag to set.

    target : None
        ignored


    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the input flags.

    References
    ----------
    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    """
    if target:
        raise NotImplementedError("target is not implemented for this function")

    fields = toSequence(field)
    df = data[fields].loc[data[fields].index_of("shared")].to_df()

    if isinstance(method, str):

        if method == "modZscore":
            MAD_series = df.subtract(df.median(axis=1), axis=0).abs().median(axis=1)
            diff_scores = (
                (0.6745 * (df.subtract(df.median(axis=1), axis=0)))
                .divide(MAD_series, axis=0)
                .abs()
            )

        elif method == "Zscore":
            diff_scores = (
                df.subtract(df.mean(axis=1), axis=0)
                .divide(df.std(axis=1), axis=0)
                .abs()
            )

        else:
            raise ValueError(method)

    else:

        try:
            stat = getattr(df, method.__name__)(axis=1)
        except AttributeError:
            stat = df.aggregate(method, axis=1)

        diff_scores = df.subtract(stat, axis=0).abs()

    mask = diff_scores > thresh
    if mask.empty:
        return data, flags

    for var in fields:
        flags[mask[var], var] = flag

    return data, flags



#!/usr/bin/env python

import pandas as pd
import numpy as np
import dios


def test_init():
    from saqc import SaQC, Flags

    arr = np.array(
        [
            [0, 1, 2],
            [0, 1, 3],
        ]
    )
    data = pd.DataFrame(arr, columns=list("abc"))
    qc = SaQC(data)

    assert isinstance(qc, SaQC)
    assert isinstance(qc._flags, Flags)
    assert isinstance(qc._data, dios.DictOfSeries)
